unit BOSCompilerU;

interface
uses
  windows,
  sysutils,
  classes,
  contnrs,
  ScannerU,
  PreProcessorU;
  
Procedure CompileBOS( filename : string;
                      EnableBASMExtentions, CompileAndDoNotAssemble, AutoEmitLastRet : boolean;
                      const AssemblerExeName, AssemblerParams, AssemblerOutputFile : string );


Type
  ECompileError = class( EScannerError );

  TCaseSensitiveStringList = class( TStringList)
  public
    function Find(const S: string; var Index: Integer): Boolean; override;
    function IndexOf(const S: string): Integer; override;
  end;

  TCompileDefine = Class
    Name : string;
    Value : string;
  end; {TCompileDefine}

  TIdentScope = class;
  TIdent = class;

  TIdentFlags = (
                 IsPiece,
                 IsLocalVar,
                 IsGlobalVar,
                 IsConst,
                 IsUnitProperty,
                 UnknownVarType
                );

  TIdentScope = class
    fScopeName : string;
    fParentScope : TIdentScope;
    fIdents : TCaseSensitiveStringList;
    fIdentsOwner : TObjectList;
    fByteCodeStream : TStream;
    SubScopeCount : integer;
    LocalVarCount : integer;
    LabelCount : integer;
  public
    ScopeIndex : integer;

    Constructor Create( const aScopeName : string; aParentScope : TIdentScope );
    Destructor Destroy; override;

    Function FindIdent( Const IdentName: string;
                        var IdentNameIsScope : boolean ) : TIdent;
    Function FindSubScope( Const ScopeName: string;
                           var ScopeNameIsIdent : boolean ) : TIdentScope;

    Function NewScope( Const ScopeName : string ) : TIdentScope;
    Function NewIdent( Const IdentName : string;
                       aFlags : TIdentFlags ) : TIdent;

    Property ScopeName : string read fScopeName;
    Property ParentScope : TIdentScope read fParentScope;
    Property ByteCodeStream : TStream read fByteCodeStream write fByteCodeStream;
  end; {TIdentScope}

  TIdent = class
  Protected
    fIdentScope : TIdentScope;
    fName : string;
    fFlags : TIdentFlags;
    fValid : boolean;
    fIndexValue : integer;
    Procedure SetIndexValue( value : integer);
  public
    // the exact meaning of this value changes depending on the ident type
    Property IndexValue : integer read fIndexValue write SetIndexValue;

    Property Valid : boolean read fValid;

    Constructor Create( aIdentScope : TIdentScope;
                        Const aName: string;
                        aFlags : TIdentFlags );


    property IdentScope : TIdentScope read fIdentScope;
    property Name : string read fName;
    property Flags : TIdentFlags read fFlags;
  end; {TIdent}

  TGobalIdentScope = class( TIdentScope )
  public
    GlobalVarCount : integer;
  end; {TGobalIdentScope}

  TBOSCompiler = class
  protected
    Text : TStringList;
    ErrorList : TStringlist;
    Scanner : TScanner;
    OutputStream : TStream;
    AsmListing : TStrings;
    PreProcessorText : TStrings;
  protected
    CompilerDefines : TCaseSensitiveStringList;

    GlobalIdentScope : TGobalIdentScope;
    CurrentIdentScope : TIdentScope;
    PieceList : TList;

    Function FindIdent( Const IdentName : string;
                        var IdentNameIsScope : boolean ) : TIdent;
    Function NewIdent( Const IdentName : string;
                       Flags : TIdentFlags ) : TIdent;
    Function FindScope( Const ScopeName : string;
                        var ScopeNameIsIdent  : boolean ) : TIdentScope;
    Function ActivateScope( Const ScopeName : string ) : TIdentScope;
    Procedure DeactivateCurrentScope;

  protected
    function DoConstFactor( Scanner : TScanner ) : Integer;
    function DoConstUnaryFactor( Scanner : TScanner ) : Integer;
    function DoConstMulTerm( Scanner : TScanner ) : Integer;
    function DoConstAddTerm( Scanner : TScanner ) : Integer;
    function DoConstShiftTerm( Scanner : TScanner ) : Integer;
    function DoConstRelationTerm( Scanner : TScanner ) : Integer;
    function DoConstEqualityTerm( Scanner : TScanner ) : Integer;
    function DoConstExpression( Scanner : TScanner ) : Integer;

    function EvaluateConstExpr( expr : string ) : Integer;
  protected
    fSkipLevels : TList;
    Procedure OnPreProcessorCommand( Sender : TPreProcessor );
    function OnExpandMacro( const MacroName : string; var NewText : string ) : boolean;
  protected
    BASMFilename : string;
    fDoOutput : boolean;
    LoopCount : Integer;
    LoopContinueLabel : string;
    UsedBreak : boolean;
    LoopBreakLabel : string;

    LocalVarList : string;
    IndentLevel : integer;
    fEnableBASMExtentions : boolean;
    fAutoEmitLastRet : boolean;
//    LastEmittedInstruction : TToken;

    Function GetPiece( var aPieceName : string ) : Integer;
    Function GetAxis( var aAxisName : string ) : Integer;
    Function GetFunc : string;
    Function GetLabel : integer;

    Procedure DoPush( value : TIdent );
    Procedure DoPop( value : TIdent );
    Procedure DoPushConst( value : Longint );


    Procedure DoGlobalDeclarations;
    Procedure DoPiece;
    Procedure DoGlobalVar;
    Procedure DoConst;
    Procedure DoFunction;
    Procedure DoStatementBlock( ForceBlock : boolean = false );
    function DoStatement : boolean;

    Function BiuldParamList( OutputLine : integer ) : integer;

    Procedure DoLocalVar;
    Procedure DoIf;
    Procedure DoWhile;
    Procedure DoDoWhile;
    Procedure DoFor;
    Procedure DoBreak;
    Procedure DoContinue;
    Procedure DoReturn;
    Procedure DoBASMInline;

    Procedure AssignmentHelper( OutputLine : integer;
                                var ResultIdent : TIdent); overload;
    Procedure DoAssignment;
    Procedure DoPreIncrement;
    Procedure DoPreDecrement;

    Procedure DoGetUnitProperty( OutputLine : integer );
    Procedure DoSetUnitProperty;
    Procedure DoSleep;
    Procedure DoTurn;
    Procedure DoWaitForTurn;
    Procedure DoMove;
    Procedure DoWaitForMove;
    Procedure DoSpin;
    Procedure DoStopSpin;
    Procedure DoHide;
    Procedure DoShow;
    Procedure DoCache;
    Procedure DoDontCache;
    Procedure DoShade;
    Procedure DoDontShade;
    Procedure DoCallScript( CanReturnValue : boolean{ = false};
                            OutputLine : integer{ = 0};
                            NoCallScriptMarker : boolean{ = false});
    Procedure DoStartScript( CanReturnValue : boolean{ = false};
                             OutputLine : integer{ = 0});
    Procedure DoSignal;
    Procedure DoSetSignalMask;
    Procedure DoExplode;
    Procedure DoEmitSfx;
    Procedure DoRand( OutputLine : integer );
    Procedure DoAttachUnit;
    Procedure DoDropUnit;

    Procedure DoFactor( OutputLine : integer );
    Procedure DoUnaryFactor( OutputLine : integer );
    Procedure DoMulTerm( OutputLine : integer );
    Procedure DoAddTerm( OutputLine : integer );
    Procedure DoShiftTerm( OutputLine : integer );
    Procedure DoRelationTerm( OutputLine : integer );
    Procedure DoEqualityTerm( OutputLine : integer );

    Procedure DoExpression( OutputLine : integer );
  protected
    Procedure AddCompilerDefine( const Name,value : string );
    Procedure DeleteCompilerDefine( const Name : string );

    function WriteMsg( const s : string ) : integer; overload;
    Procedure WriteMsg( index : integer; const s : string ) ; overload;
    function WriteErrorMsg( const s : string ) : integer;
    function WriteStatusMsg( const s : string ) : integer;

    procedure WriteBeginStatusMsg( var index : integer; const s : string );
    Function WritePartialStatusMsg( const s : string ) : integer; overload;
    Procedure WritePartialStatusMsg( index : integer; const s : string ); overload;

    Procedure Compile;
  public
    Constructor Create( aText : TStringlist;
                        aOutputStream : TStream;
                        aAsmListing : TStrings;
                        aPreProcessorText : TStrings;
                        aEnableBASMExtentions,aAutoEmitLastRet : boolean;
                        aBASMFilename : string);
    Destructor destroy; override;

    Property AutoEmitLastRet : boolean read fAutoEmitLastRet write fAutoEmitLastRet;
    Property EnableBASMExtentions : boolean read fEnableBASMExtentions write fEnableBASMExtentions;
    Property DoOutput : boolean read fDoOutput write fDoOutput;
  end; {TBOSCompiler}


Type
  TUnitProperty =record
    Key : Longint;
    Name : string;
    Flags : set of ( SetAble, ArgIsPiece, ArgIsUnitID, IsBool );
    // max of 5
    ArgCount : 0..4;
  end; {TUnitProperty}  
// Indices for set/get value
const
  // set Or get
  Activation         =  1;
  // set Or get
  StandingMoveOrders =  2;
  // set Or get
  StandingFireOrders =  3;
  // get ( 0-100% )
  Health             =  4;
  // set Or get
  InBuildStance      =  5;
  // set Or get ( used by misc. special case missions like transport ships )
  Busy               =  6;
  // get
  Piece_XZ           =  7;
  // get
  Piece_Y            =  8;
  // get
  Unit_XZ            =  9;
  // get
  Unit_Y             = 10;
  // get
  Unit_Height        = 11;
  // get atan of packed x,z coords
  XZ_aTan            = 12;
  // get hypot of packed x,z coords
  XZ_Hypot           = 13;
  // get ordinary two-parameter atan
  aTan               = 14;
  // get ordinary two-parameter hypot
  Hypot              = 15;
  // get
  Ground_Height      = 16;
  // get 0 = unit is built and ready, 1-100 = How much is left to build
  Build_Percent_Left = 17;
  // set Or get ( change which plots we occupy when building opens and closes )
  Yard_Open          = 18;
  // set Or get ( ask other units to clear the area )
  Bugger_Off         = 19;
  // set Or get
  Armored            = 20;

  // various unit properties
  UnitProperties : array [  1..20 ] of TUnitProperty = ( 
                 ( key : Activation;            Name : 'ACTIVATION';
                  Flags : [  SetAble,IsBool ];  ArgCount : 0 ),
                 ( key : StandingMoveOrders;    Name : 'STANDINGMOVEORDERS';
                  Flags : [  SetAble ];         ArgCount : 0 ),
                 ( key : StandingFireOrders;    Name : 'STANDINGFIREORDERS';
                  Flags : [  SetAble ];         ArgCount : 0 ),
                 ( key : Health;                Name : 'HEALTH';
                  Flags : [ ];                  ArgCount : 0 ),
                 ( key : InBuildStance;         Name : 'INBUILDSTANCE';
                  Flags : [  SetAble ];         ArgCount : 0 ),
                 ( key : Busy;                  Name : 'BUSY';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ),
                 ( key : Piece_XZ;              Name : 'PIECE_XZ';
                  Flags : [  ArgIsPiece ];      ArgCount : 1 ),
                 ( key : Piece_Y;               Name : 'PIECE_Y';
                  Flags : [  ArgIsPiece ];      ArgCount : 1 ),
                 ( key : Unit_XZ;               Name : 'UNIT_XZ';
                  Flags : [  ArgIsUnitID ];     ArgCount : 1 ),
                 ( key : Unit_Y;                Name : 'UNIT_Y';
                  Flags : [  ArgIsUnitID ];     ArgCount : 1 ),
                 ( key : Unit_Height;           Name : 'UNIT_Height';
                  Flags : [  ArgIsUnitID ];     ArgCount : 1 ),
                 ( key : XZ_aTan;               Name : 'XZ_ATAN';
                  Flags : [ ];                  ArgCount : 1 ),
                 ( key : XZ_Hypot;              Name : 'XZ_HYPOT';
                  Flags : [ ];                  ArgCount : 1 ),
                 ( key : aTan;                  Name : 'ATAN';
                  Flags : [ ];                  ArgCount : 2 ),
                 ( key : Hypot;                 Name : 'HYPOT';
                  Flags : [ ];                  ArgCount : 2 ),
                 ( key : Ground_Height;         Name : 'GROUND_HEIGHT';
                  Flags : [ ];                  ArgCount : 1 ),
                 ( key : Build_Percent_Left;    Name : 'Build_Percent_Left';
                  Flags : [ ];                  ArgCount : 0 ),
                 ( key : Yard_Open;             Name : 'YARD_OPEN';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ),
                 ( key : Bugger_Off;            Name : 'BUGGER_OFF';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ),
                 ( key : Armored;               Name : 'ARMORED';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ) );
                    
implementation
uses
  shellapi,
  FileHandlingU,
  Cob_Format;

// -----------------------------------------------------------------------------
//   TCaseSensitiveStringList
// -----------------------------------------------------------------------------

function TCaseSensitiveStringList.IndexOf(const S: string): Integer;
begin
if not Sorted then
  begin
  for Result := 0 to GetCount - 1 do
    if AnsiCompareStr(Get(Result), S) = 0 then Exit;
  Result := -1;
  end
else if not Find(S, Result) then Result := -1;
end;

function TCaseSensitiveStringList.Find(const S: string; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := AnsiCompareStr(self[I], S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        {if list.Duplicates <> dupAccept then }L := I;
      end;
    end;
  end;
  Index := L;
end; {Find}

// -----------------------------------------------------------------------------

Procedure Exec( const Filename, args : string; const DoRedirect : string );
var
  si : TStartupInfo;
  pi : TProcessInformation;
begin
writeln( 'Attempting to invoke '+Filename);
fillchar(si, sizeof( si ), 0 );
si.cb := sizeof( si );
si.dwFlags := STARTF_USESTDHANDLES;
si.hStdInput := GetStdHandle( STD_INPUT_HANDLE );
si.hStdOutput := GetStdHandle( STD_OUTPUT_HANDLE );
si.hStdError := GetStdHandle( STD_ERROR_HANDLE );
fillchar(pi, sizeof( pi ), 0 );
if CreateProcess( pchar( Filename ), pchar( ' '+args ),
                  nil, nil, true, 0, nil, nil, si, pi) then
  begin
  WaitForSingleObject(pi.hProcess, INFINITE);
  end
else
  begin
  writeln( 'Unable to invoke '+Filename);
  writeln( 'Reason: '+SysErrorMessage( GetLastError ) );
  end;
end; {Exec}

Procedure CompileBOS( filename : string;
                      EnableBASMExtentions, CompileAndDoNotAssemble, AutoEmitLastRet : boolean;
                      const AssemblerExeName, AssemblerParams, AssemblerOutputFile : string );
var
  BOSCompiler : TBOSCompiler;
  Stream : TMemoryStream;
  AsmListing, PreProcessorText : TStringlist;
  OutputFileName : string;
  BASMFileName : string;
  PreProcessorFilename : string;
  ErrorFileName : string;
begin
Stream := nil;
AsmListing := nil;
PreProcessorText := nil;
BOSCompiler := nil;
try
  AsmListing := TStringlist.create;
  PreProcessorText := TStringlist.create;
  Stream := TMemoryStream.create;
  filename := FindFile( filename );
  BASMFileName := ChangeFileExt( filename, '.basm');
  PreProcessorFilename := ChangeFileExt( filename, '.prebasm');
  OutputFileName := ChangeFileExt( filename, '.cob');
  ErrorFileName := ChangeFileExt( filename, '_BASM_error.txt');
  writeln( 'Source: '+filename );
  writeln( 'Dest: '+BASMFileName );
  writeln( 'Preprocessor Output: '+PreProcessorFilename );
  writeln( 'BASM error log: '+ErrorFileName);
  try
    BOSCompiler := TBOSCompiler.Create( Loadfile( filename ),
                                        Stream, AsmListing, PreProcessorText,
                                        EnableBASMExtentions,AutoEmitLastRet,BASMFileName );
  finally
    if FileExists( BASMFileName ) then
      DeleteFile( BASMFileName );
    AsmListing.SaveToFile( BASMFileName );
    if FileExists( PreProcessorFilename ) then
      DeleteFile( PreProcessorFilename );
    PreProcessorText.SaveToFile( PreProcessorFilename );
    if BOSCompiler.DoOutput and not CompileAndDoNotAssemble then
      begin
      writeln( 'Assembling BASM to COB' );  
      if FileExists( OutputFileName ) then
        DeleteFile( OutputFileName );
      BASMFileName := AnsiQuotedStr(BASMFileName, '"' );
      OutputFileName := AnsiQuotedStr(OutputFileName, '"' );
      Exec( AssemblerExeName, BASMFileName +' '+ OutputFileName +' ' +
                              AssemblerParams, AssemblerOutputFile);
      If FileExists( 'errors.txt' ) then
        begin
        // and error occured while making the COB, move the 'error.txt' file to
        // the same directory as the rest of the output.
        AsmListing.LoadFromFile( 'errors.txt' );
        DeleteFile( 'errors.txt' );
        if FileExists( ErrorFileName ) then
          DeleteFile( ErrorFileName );
        AsmListing.SaveToFile( ErrorFileName );
        end;
      end
    else if not BOSCompiler.DoOutput then
      begin
      writeln( 'Errors in source' );
      writeln( BOSCompiler.ErrorList.text );
      end;
    writeln( 'Finished Compiling!' );
  end;
finally
  Stream.free;
  AsmListing.free;
  PreProcessorText.free;
  BOSCompiler.free;
end;
end; {CompileBOS}

// -----------------------------------------------------------------------------
//   TIdentScope
// -----------------------------------------------------------------------------

Constructor TIdentScope.Create( const aScopeName : string;
                                aParentScope : TIdentScope );
begin
inherited create;
assert( aScopeName <> '' );
fScopeName := aScopeName;
fParentScope := aParentScope;
fIdents := TCaseSensitiveStringList.create;
fIdents.Sorted := true;
fIdents.Duplicates := dupError;
fIdentsOwner := TObjectList.create( true );
fByteCodeStream := TMemoryStream.create;
end; {Create}

Destructor TIdentScope.Destroy;
begin
FreeAndNil( fIdents );
FreeAndNil( fIdentsOwner );
FreeAndNil( fByteCodeStream );
inherited;
end; {Destroy}

Function TIdentScope.FindIdent( Const IdentName: string;
                                var IdentNameIsScope : boolean ) : TIdent;
var
  obj : TObject;
  index : integer;
begin
assert( self <> nil );
assert( fIdents <> nil );
result := nil;
IdentNameIsScope := false;
if IdentName = '' then exit;
if fIdents.Find( IdentName, index ) then
  begin
  obj := fIdents.Objects[ index];
  if obj is TIdent then
    result := TIdent( obj )
  else if obj is TIdentScope then
    IdentNameIsScope := true;
  end
else if ParentScope <> nil then
  result := ParentScope.FindIdent( IdentName, IdentNameIsScope );
end; {FindIdent}

Function TIdentScope.FindSubScope( Const ScopeName: string;
                                   var ScopeNameIsIdent : boolean ) : TIdentScope;
var
  obj : TObject;
  index : integer;
begin
assert( self <> nil );
assert( fIdents <> nil );
result := nil;
ScopeNameIsIdent := false;
if ScopeName = '' then exit;
if fIdents.Find( ScopeName, index ) then
  begin
  obj := fIdents.Objects[ index];
  if obj is TIdentScope then
    result := TIdentScope( obj )
  else if obj is TIdent then
    ScopeNameIsIdent := true;
  end;
if ( result = nil ) and ( ParentScope <> nil ) then
  Result := ParentScope.FindSubScope( ScopeName, ScopeNameIsIdent );
end; {FindSubScope}

Function TIdentScope.NewScope( Const ScopeName : string ) : TIdentScope;
begin
assert( self <> nil );
assert( fIdents <> nil );
assert( fIdentsOwner <> nil );
if fIdents.IndexOf( ScopeName ) = -1 then
  begin
  result := TIdentScope.create( ScopeName, self );
  fIdentsOwner.add( result );
  fIdents.AddObject( ScopeName ,result );
  result.ScopeIndex := SubScopeCount;
  inc( SubScopeCount );
  end
else raise ECompileError.create( 'Redeclared Identifier: "' + ScopeName + '"' );
end; {NewScope}

Function TIdentScope.NewIdent( Const IdentName : string;
                               aFlags : TIdentFlags ) : TIdent;
begin
assert( self <> nil );
assert( fIdents <> nil );
assert( fIdentsOwner <> nil );
if fIdents.IndexOf( ScopeName ) = -1 then
  begin
  result := TIdent.create( self, IdentName, aFlags );
  fIdentsOwner.add( result );
  fIdents.AddObject( IdentName ,result );
  end
else  raise ECompileError.create( 'Redeclared Identifier: "' + IdentName + '"' );
end; {NewIdent}

// -----------------------------------------------------------------------------
//   TIdent
// -----------------------------------------------------------------------------

Constructor TIdent.Create( aIdentScope : TIdentScope;
                           Const aName: string;
                           aFlags : TIdentFlags );
begin
assert( aIdentScope <> nil );
assert( aName <> '' );
fIdentScope := aIdentScope;
fName := aName;
fFlags := aFlags;
IndexValue := -1;
end; {Create}

Procedure TIdent.SetIndexValue( value : integer);
begin
fIndexValue := value;
fValid := true;
end; {SetIndexValue}

// -----------------------------------------------------------------------------
//   TBOSCompiler
// -----------------------------------------------------------------------------

Constructor TBOSCompiler.Create(  aText : TStringlist;
                                  aOutputStream : TStream;
                                  aAsmListing,aPreProcessorText : TStrings;
                                  aEnableBASMExtentions, aAutoEmitLastRet : boolean;
                                  aBASMFilename : string);
begin
Inherited Create;
Text := aText;
OutputStream := aOutputStream;
AsmListing := aAsmListing;
PreProcessorText := aPreProcessorText;
BASMFilename := aBASMFilename;
assert( Text <> nil );
assert( OutputStream <> nil );
assert( AsmListing <> nil );
assert( PreProcessorText <> nil );
EnableBASMExtentions := aEnableBASMExtentions;
AutoEmitLastRet := aAutoEmitLastRet;
CompilerDefines := TCaseSensitiveStringList.create;
CompilerDefines.Sorted := true;
CompilerDefines.Duplicates := dupAccept;
PieceList := TList.create;
GlobalIdentScope := TGobalIdentScope.create( 'global scope',nil );
CurrentIdentScope := GlobalIdentScope;
fSkipLevels := TList.create;
ErrorList := TStringlist.Create;
Compile;
//aText.Text := Scanner.Text;
//aText.SaveToFile('test2.txt');
end; {Create}

Destructor TBOSCompiler.Destroy;
begin
Scanner.free;
GlobalIdentScope.free;
CompilerDefines.free;
PieceList.free;
ErrorList.free;
Text.free;
Inherited Destroy;
end; {Destroy}

Procedure TBOSCompiler.AddCompilerDefine( const Name,value : string );
var
  CompileDefine : TCompileDefine;
begin
assert( self <> nil );
assert( CompilerDefines <> nil );
CompileDefine := TCompileDefine.create;
try
  CompileDefine.Name := Name;
  CompileDefine.Value := Trim( value );
  CompilerDefines.AddObject( CompileDefine.Name , CompileDefine );
//  WriteMsg( '#Define ' + Name + ' ' + value);
except
  CompilerDefines.free;
  raise;
end;
end; {AddCompilerDefine}

Procedure TBOSCompiler.DeleteCompilerDefine( const Name : string );
var
  CompileDefine : TCompileDefine;
  index : integer;
begin
assert( self <> nil );
assert( CompilerDefines <> nil );
if CompilerDefines.find( Name, index ) then
  begin
  CompileDefine := CompilerDefines.objects[  index ] as TCompileDefine;
  CompilerDefines.delete( index );
  CompileDefine.Free;
  end;
end; {DeleteCompilerDefine}

// -----------------------------------------------------------------------------

function TBOSCompiler.OnExpandMacro( const MacroName : string;
                                     var NewText : string ) : boolean;
var
  index : integer;
begin
assert( self <> nil );
assert( CompilerDefines <> nil );
Result := CompilerDefines.Find( MacroName, index );
if Result then
  NewText := TCompileDefine( CompilerDefines.Objects[  index ] ).Value;
end; {OnExpandMacro}


Function CopyFrom(S : string; Index, Count: Integer): string;
begin
result := copy(S,index,count);
end;

Procedure TBOSCompiler.OnPreProcessorCommand( Sender : TPreProcessor  );
var
  startIndex : integer;
  includefile : TStringlist;
  Value : string;
  SkipLevel : integer;
  Skipping : boolean;
begin
CopyFrom('a',1,1);
assert( self <> nil );
assert( CompilerDefines <> nil );
assert( Sender <> nil );
SkipLevel := fSkipLevels.count-1;
Skipping := (SkipLevel > 0) and (fSkipLevels[SkipLevel] <> nil);
if Sender.Token = CSTI_Identifier then
  begin
  startIndex := Sender.Position;
  Value := Sender.Value;
  if Value = 'include' then
    begin
    Sender.GetToken;
    if not Skipping and (Sender.Token = CSTI_String) then
      begin
      includefile := Loadfile( FindFile( Sender.Value ) );
      try
        Sender.SkipToNewLine;
        Sender.RemoveString( startIndex, Sender.NextCharIndex - startIndex-1 );
        Sender.NextCharIndex := startIndex;
        Sender.InsertString( startIndex, includefile.text, false );
        Sender.Token := CSTINT_WhiteSpace;
        Sender.SkipWhite;
      finally
        includefile.free;
      end;
//      PreProcessorText.Text := sender.Text;
//      PreProcessorText.SaveToFile('test.txt');
      end;
    end
  else if Value = 'define' then
    begin
    Sender.GetToken;
    if not Skipping and (Sender.Token = CSTI_Identifier) then
      AddCompilerDefine( Sender.Value, Sender.TillNewLine );
    Sender.SkipToNewLine;
    Sender.SkipNewLine;
    Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
    Sender.SkipWhite;
    end
  else if Value = 'undef' then
    begin
    Sender.GetToken;
    if not Skipping and ( Sender.Token = CSTI_Identifier ) then
      begin
      DeleteCompilerDefine( Sender.Value );
      Sender.SkipToNewLine;
      Sender.SkipNewLine;
      Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
      Sender.SkipWhite;
      end;
    end
  else if Value = 'if' then
    begin
    Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
//    asm int 3 end;
    StartIndex := Sender.NextCharIndex-1;
    value := Sender.TillNewLine;
    Sender.SkipToNewLine;
    Sender.SkipNewLine;
    Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
    if not Skipping then
      begin
      if (EvaluateConstExpr( value ) = 0) then
        begin
        SkipLevel := fSkipLevels.add(Pointer(1));
        While ( Sender.Token <> CSTI_EOF ) and
              ( fSkipLevels.count > SkipLevel ) and
              ( SkipLevel >= 0 ) and
              ( fSkipLevels[SkipLevel] <> nil ) do
          begin
          StartIndex := Sender.NextCharIndex-1;
          Sender.GetToken;
          Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
          end;
        end
      else
        fSkipLevels.add( nil )
      end
    else
      fSkipLevels.add( Pointer(1) );
    end
  else if Value = 'ifdef' then
    begin
    Sender.GetToken;
    if ( Sender.Token = CSTI_Identifier ) then
      begin
      Value := Sender.Value;
      Sender.SkipToNewLine;
      Sender.SkipNewLine;
      Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
      Sender.SkipWhite;
      if not Skipping then
        begin
        if CompilerDefines.IndexOf( Value ) = -1 then
          begin
          SkipLevel := fSkipLevels.add(Pointer(1));
          While ( Sender.Token <> CSTI_EOF ) and
                ( SkipLevel >= 0 ) and
                ( fSkipLevels.count > SkipLevel ) and
                ( fSkipLevels[SkipLevel] <> nil ) do
            begin
            StartIndex := Sender.NextCharIndex-1;
            Sender.GetToken;
            Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
            end;
          end
        else
          fSkipLevels.add( nil );
        end
      else
        fSkipLevels.add( Pointer(1) );
      end;
    end
  else if Value = 'ifndef' then
    begin
    Sender.GetToken;
    if Sender.Token = CSTI_Identifier then
      begin
      Value := Sender.Value;
      Sender.SkipToNewLine;
      Sender.SkipNewLine;
      Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
      Sender.SkipWhite;
      if not Skipping then
        begin
        if CompilerDefines.IndexOf( Value ) <> -1 then
          begin
          SkipLevel := fSkipLevels.add(Pointer(1));
          While ( Sender.Token <> CSTI_EOF ) and
                ( SkipLevel >= 0 ) and
                ( fSkipLevels.count > SkipLevel ) and
                ( fSkipLevels[SkipLevel] <> nil ) do
            begin
            StartIndex := Sender.NextCharIndex-1;
            Sender.GetToken;
            Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
            end;
          end
        else
          fSkipLevels.add( nil );
        end
      else
        fSkipLevels.add( Pointer(1) );
      end;
    end
  else if Value = 'else' then
    begin
    Sender.SkipToNewLine;
    Sender.SkipNewLine;
    Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
    Sender.SkipWhite;
    if (SkipLevel >= 0 )and ( fSkipLevels.count > SkipLevel) then
      begin
      if (fSkipLevels[ SkipLevel ] = nil) then
        begin // need to skip
        fSkipLevels[ SkipLevel ] := Pointer(1);
        While ( Sender.Token <> CSTI_EOF ) and
              ( SkipLevel >= 0 ) and           
              ( fSkipLevels.count > SkipLevel ) and
              ( fSkipLevels[SkipLevel] <> nil ) do
          begin
          StartIndex := Sender.NextCharIndex-1;
          Sender.GetToken;
          Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
          end;
        end
      else
        fSkipLevels[ SkipLevel ] := nil;
      end
    else
      raise ECompileError.create( 'Conditional define error, #else' );
    end
  else if Value = 'endif' then
    begin
    if (SkipLevel >= 0 )and ( fSkipLevels.count > SkipLevel) then
      fSkipLevels.Delete( SkipLevel )
    else
      raise ECompileError.create( 'Conditional define error, #endif' );
    Sender.SkipToNewLine;
    Sender.SkipNewLine;
    Sender.RemoveString( StartIndex, Sender.NextCharIndex - StartIndex-1 );
    Sender.SkipWhite;
    end;
  end;
end; {OnPreProcessorCommand}

// -----------------------------------------------------------------------------

Procedure TBOSCompiler.WriteMsg( index : integer; const s : string ) ;
begin
assert( self <> nil );
//writeln( s );
AsmListing.Insert( index, s );
end; {WriteMsg}

function TBOSCompiler.WriteMsg( const s : string ) : integer ;
begin
assert( self <> nil );
//writeln( s );
Result := AsmListing.add( StringOfChar(' ',IndentLevel) + s );
end; {WriteMsg}

function TBOSCompiler.WriteStatusMsg( const s : string ) : integer;
begin
assert( self <> nil );
result := WriteMsg( '// ' + s );
end; {WriteStatusMsg}

Procedure TBOSCompiler.WriteBeginStatusMsg(  var index : integer; const s : string );
begin
assert( self <> nil );
index := AsmListing.add( StringOfChar(' ',IndentLevel) + '// ' + s );
end; {WriteStatusMsg}

Function TBOSCompiler.WritePartialStatusMsg( const s : string ) : integer;
begin
assert( self <> nil );
result := AsmListing.count-1;
AsmListing[ result ] := AsmListing[ result ] + s;
end; {WriteStatusMsg}

Procedure TBOSCompiler.WritePartialStatusMsg( index : integer; const s : string );
begin
assert( self <> nil );
AsmListing[ index ] := AsmListing[ index ] + s;
end; {WriteStatusMsg}

function TBOSCompiler.WriteErrorMsg( const s : string ) : integer;
begin
DoOutput := false;
result := WriteStatusMsg( 'Error: ' + s );
ErrorList.add( '[Error] '+BASMFilename+'('+IntTostr( 1+result )+'): ' +s);
end; {WriteErrorMsg}

// -----------------------------------------------------------------------------

// Parse and Translate a Math Factor
function TBOSCompiler.DoConstFactor( Scanner : TScanner ) : integer;
var
  Ident : TIdent;
  IntValue : Int64;
  IdentNameIsScope : boolean;
begin
case Scanner. token of
  CSTI_OpenRound :
    begin
    Scanner.Match( CSTI_OpenRound );
    Result := DoConstExpression( Scanner );
    Scanner.Match( CSTI_CloseRound );
    end;
  CSTI_Identifier :
    begin
    Ident := FindIdent( Scanner.Value, IdentNameIsScope );
    if Ident = nil then
       begin  // referencing a non-existant variable
       WriteErrorMsg( 'Undeclaired identifier, ' + Scanner.Value );
       Ident := NewIdent( Scanner.Value, UnknownVarType );
       end;
    if Ident.Flags in [IsLocalVar, IsGlobalVar ] then
      Raise ECompileError.create( 'Constant expression expected, but found a varaible '+Ident.Name );
    if not Ident.Valid then
      begin
      WriteErrorMsg( 'Undeclaired identifier, ' + Ident.Name );
      Raise ECompileError.create( 'Undeclaired identifier, ' + Ident.Name );
      end;
    result := Ident.IndexValue;
    Scanner.GetToken;
    end;
  CSTI_Integer, CSTI_HexInt :
    begin
    IntValue := StrToInt64( Scanner.Value );
    if IntValue > High(Longword) then
      IntValue := IntValue mod High(Longword);
    result := IntValue;
    Scanner.GetToken;
    end;
  CSTI_Real :
    begin
    WriteErrorMsg( 'Floatpoint values are not permitted in arbitrary locations ' + Scanner.Value );
    Scanner.GetToken;
    result := 0;
    end
  else Raise ECompileError.create( 'Unknown token: ' + TokenName( Scanner.Token ) );
end;    
end; {DoConstFactor}

function TBOSCompiler.DoConstUnaryFactor( Scanner : TScanner ) : integer;
begin
case Scanner.Token of
  CSTI_BitwiseNot :
    begin
    Scanner.GetToken;
    result := not DoConstFactor( Scanner );
    end;
  CSTI_LogicalNot :
    begin
    Scanner.GetToken;
    result := Integer( not Boolean( DoConstFactor( Scanner ) ) );
    end;
  CSTI_Minus :
    begin
    Scanner.GetToken;
    result := - DoConstFactor( Scanner );
    end;
else result := DoConstFactor( Scanner );
end;
end; {DoConstUnaryFactor}

function TBOSCompiler.DoConstMulTerm( Scanner : TScanner ) : integer;
begin
result := DoConstUnaryFactor( Scanner );
while true do
case Scanner.Token of
  CSTI_Multiply :
    begin
    Scanner.Match( CSTI_Multiply );
    result := result * DoConstUnaryFactor( Scanner );
    end;
  CSTI_Divide :
    begin
    Scanner.Match( CSTI_Divide );
    result := result div DoConstUnaryFactor( Scanner );
    end;
  else break;
end;
end; {DoMulTerm}

function TBOSCompiler.DoConstAddTerm( Scanner : TScanner ) : integer;
begin
result := DoConstMulTerm( Scanner );
while true do
case Scanner.Token of
  CSTI_Plus :
    begin
    Scanner.Match( CSTI_Plus );
    result := result + DoConstMulTerm( Scanner );
    end;
  CSTI_Minus :
    begin
    Scanner.Match( CSTI_Minus );
    result := result - DoConstMulTerm( Scanner );
    end;
  else break;
end;
end; {DoConstAddTerm}

function sar(a : integer; b : byte) : integer;
asm
  mov cl, b;
  sar a, cl;
end;

function TBOSCompiler.DoConstShiftTerm( Scanner : TScanner ) : integer;
begin
Result := DoConstAddTerm( Scanner );
if EnableBASMExtentions then
while true do
case Scanner.Token of
  CSTI_ShiftLeft :
    begin
    Scanner.Match( CSTI_ShiftLeft );
    Result := Result shl DoConstAddTerm( Scanner );
    end;
  CSTI_ShiftRight :
    begin
    Scanner.Match( CSTI_ShiftRight );
    Result := sar( Result, DoConstAddTerm( Scanner ) );
    end;
  else break;
end;
end; {DoConstShiftTerm}

function TBOSCompiler.DoConstRelationTerm( Scanner : TScanner ) : integer;
begin
result := DoConstShiftTerm( Scanner );
case Scanner.Token of
  CSTI_Greater :
    begin
    Scanner.Match( CSTI_Greater );
    result := Integer( result > DoConstShiftTerm( Scanner ) );
    end;
  CSTI_GreaterEqual :
    begin
    Scanner.Match( CSTI_GreaterEqual );
    result := Integer( result >= DoConstShiftTerm( Scanner ) );
    end;
  CSTI_Less :
    begin
    Scanner.Match( CSTI_Less );
    result := Integer( result < DoConstShiftTerm( Scanner ) );
    end;
  CSTI_LessEqual :
    begin
    Scanner.Match( CSTI_LessEqual );
    result := Integer( result <= DoConstShiftTerm( Scanner ) );
    end;
end;
end; {DoConstRelationTerm}

function TBOSCompiler.DoConstEqualityTerm( Scanner : TScanner ) : integer;
begin
result := DoConstRelationTerm( Scanner );
case Scanner.Token of
  CSTI_Equal :
    begin
    Scanner.Match( CSTI_Equal );
    result := Integer( result = DoConstRelationTerm( Scanner ) );
    end;
  CSTI_NotEqual :
    begin
    Scanner.Match( CSTI_Equal );
    result := Integer( result <> DoConstRelationTerm( Scanner ) );
    end;
end;
end; {DoConstEqualityTerm}

function TBOSCompiler.DoConstExpression( Scanner : TScanner ) : integer;
begin
result := DoConstEqualityTerm( Scanner );
while true do
  begin
  if Scanner.token = CSTI_EOF then
    break;
  if Scanner.IfMatch( CSTI_LogicalOr ) then
    result := Integer( boolean( result ) or boolean( DoConstEqualityTerm( Scanner ) ) )
  else if Scanner.IfMatch( CSTI_LogicalXor ) then
    result := Integer( boolean( result ) xor boolean( DoConstEqualityTerm( Scanner  ) ) )
  else if Scanner.IfMatch( CSTI_LogicalAnd ) then
    result := Integer( boolean( result ) and boolean( DoConstEqualityTerm( Scanner  ) ) )
  else if Scanner.IfMatch( CSTI_BitwiseOr ) then
    result := result or DoConstEqualityTerm( Scanner  )
  else if Scanner.IfMatch( CSTI_BitwiseXor ) then
    result := result xor DoConstEqualityTerm( Scanner  )
  else if Scanner.IfMatch( CSTI_BitwiseAnd ) then
    result := result and DoConstEqualityTerm( Scanner  )
  else break;
  end; 
end; {DoConstExpression}

function TBOSCompiler.EvaluateConstExpr( expr : string ) : Integer;
var
  aScanner : TScanner;
  apreprocessor : TPreProcessor;
begin
apreprocessor := TPreProcessor.create( expr, nil, OnExpandMacro );
try
  While apreprocessor.token <> CSTI_EOF do
    apreprocessor.GetToken;
finally
  expr := apreprocessor.Text;
  apreprocessor.free;
end;
aScanner := TScanner.create( expr );
try
  Result := DoConstExpression( aScanner );
finally
  aScanner.free;
end;
end; {EvaluateConstExpr}

// -----------------------------------------------------------------------------

Function TBOSCompiler.FindIdent( Const IdentName : string;
                                 var IdentNameIsScope : boolean ) : TIdent;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
result := CurrentIdentScope.FindIdent( IdentName, IdentNameIsScope )
end; {FindIdent}

Function TBOSCompiler.NewIdent( Const IdentName : string;
                                Flags : TIdentFlags ) : TIdent;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
result := CurrentIdentScope.NewIdent( IdentName, Flags );
if Flags = IsPiece then
  result.IndexValue := PieceList.add( result )
else if Flags = IsLocalVar then
  begin
  result.IndexValue := CurrentIdentScope.LocalVarCount;
  inc( CurrentIdentScope.LocalVarCount );
  end
else if Flags = IsGlobalVar then
  begin
  result.IndexValue := GlobalIdentScope.GlobalVarCount;
  inc( GlobalIdentScope.GlobalVarCount );
  end
end; {NewIdent}

Function TBOSCompiler.FindScope( Const ScopeName : string;
                                 var ScopeNameIsIdent  : boolean ) : TIdentScope;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
Result := CurrentIdentScope.FindSubScope( ScopeName, ScopeNameIsIdent );
end; {FindScope}

Function TBOSCompiler.ActivateScope( Const ScopeName : string ) : TIdentScope;
var
  ScopeNameIsIdent : boolean;
begin
assert( self <> nil );
Result := FindScope( ScopeName, ScopeNameIsIdent );
if Result = nil then
  begin
  if ScopeNameIsIdent then
    Raise ECompileError.create( 'Function name "'+ScopeName+'" is used by a variable ' );
  result := CurrentIdentScope.NewScope( ScopeName );
  if Result = nil then
    Raise ECompileError.create( 'Error: Unable to activate scope:' + ScopeName );
  end;
if result <> nil then
  CurrentIdentScope := result;
end; {ActivateScope}

Procedure TBOSCompiler.DeactivateCurrentScope;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
if CurrentIdentScope.ParentScope <> nil then
  CurrentIdentScope := CurrentIdentScope.ParentScope
else
  CurrentIdentScope := GlobalIdentScope;
end; {DeactivateCurrentScope}

// -----------------------------------------------------------------------------

Procedure TBOSCompiler.Compile;
var
  StartTime,EndTime,freq : Int64;
  CompileTime : Extended;
  PreProcessor : TPreProcessor;
  index : integer;
begin
assert( self <> nil );
ErrorList.clear;
QueryPerformanceFrequency( freq );
QueryPerformanceCounter( StartTime );
try try
  // do pre-processing
  PreProcessor := nil;
  try
    PreProcessor := TPreProcessor.create( Text.Text, OnPreProcessorCommand, OnExpandMacro );
    AddCompilerDefine( 'TRUE','1' );
    AddCompilerDefine( 'FALSE','0' );
    AddCompilerDefine( 'true','1' );
    AddCompilerDefine( 'false','0' );
    AddCompilerDefine( 'x-axis','0' );
    AddCompilerDefine( 'y-axis','1' );
    AddCompilerDefine( 'z-axis','2' );
    {
    for index := low( UnitProperties ) to high( UnitProperties ) do
      AddCompilerDefine( AnsiUpperCase( UnitProperties[ index].Name ),
                         IntToStr( index ) );
    }
    While PreProcessor.token <> CSTI_EOF do
      PreProcessor.GetToken;
  finally
    if PreProcessor <> nil then
      PreProcessorText.Text := PreProcessor.Text;
    FreeAndNil(fSkipLevels);
    PreProcessor.free;
  end;
  // construct the lecican scanner
  Scanner := TScanner.create( PreProcessorText.Text );
  // do the parsing
  LoopCount := 0;
  LoopContinueLabel := '';
  LoopBreakLabel := '';
  UsedBreak := false;
  DoOutput := true;
  IndentLevel := 0;
  // parse the unit script
  DoGlobalDeclarations;
  // create the script output
except
  on E : EScannerError do
    begin
    WriteErrorMsg( '' );
    WriteStatusMsg( '********************' );
    if Scanner <> nil then
      begin
      WriteStatusMsg( 'Error parsing Token' );
      WriteStatusMsg( 'Error message : ''' + E.message + '''' );
      WriteStatusMsg( 'Token Name : "' + TokenName( Scanner.Token ) +'"' );
      WriteStatusMsg( 'Token Value : "' + Scanner.value + '"');
      WriteStatusMsg( 'Token Position : ' + Inttostr( Scanner.Position ) );
      end
    else
      begin
      WriteStatusMsg( 'Unknown error' );
      WriteStatusMsg( 'Error message : ''' + E.message + '''' );
      end;
    WriteStatusMsg( '********************' );
    end;
  On E : Exception do
    begin
    WriteErrorMsg( '' );
    WriteStatusMsg( 'Unknown error' );
    WriteStatusMsg( 'Error message : ''' + E.message + '''' );
    end;
end;
finally
  QueryPerformanceCounter( EndTime );
  CompileTime := ( EndTime - StartTime ) / freq;
  WriteStatusMsg( 'Compile time  : ' + FloatToStrF( CompileTime,ffExponent,2,0 ) + ' seconds' );
  if not DoOutput then
    begin
    WriteStatusMsg( 'Errors in source code' );
    for index := 0 to ErrorList.count-1 do
      WriteStatusMsg( ErrorList[index] );
    end;
end;
end; {Compile}

Function TBOSCompiler.GetPiece( var aPieceName : string ) : Integer;
var
  ident : TIdent;
  IdentNameIsScope : boolean;
begin
Result := -1;
aPieceName := Scanner.Value;
if Scanner.Token in [ CSTI_Integer,CSTI_HexInt] then
  Result := StrToInt( Scanner.Value )
else if Scanner.Token = CSTI_Identifier then
  begin
  ident := FindIdent( Scanner.Value, IdentNameIsScope );
  if Ident = nil then
     begin
     WriteErrorMsg( 'Undeclaired identifier, ' + Scanner.Value );
     Ident := NewIdent( Scanner.Value, UnknownVarType );
     end;
  assert( Ident <> nil );
  if ident.Flags = IsPiece then
    result := ident.IndexValue
  else if ident.Flags <> UnknownVarType then
    WriteErrorMsg( 'Expecting piece name, but found a unknown variable type' )
  end
else
  WriteErrorMsg( 'Expecting piece name or index, but found a ' + TokenName( Scanner.Token ) );
Scanner.GetToken;
end; {GetPiece}

Function TBOSCompiler.GetAxis( var aAxisName : string ) : Integer;
begin
aAxisName := Scanner.Value;
if Scanner.Token in [ CSTI_Integer,CSTI_HexInt] then
  Result := StrToInt( Scanner.Value )
else
  begin
  result := -1;
  WriteErrorMsg( 'Expecting axis name or value, but found a ' + TokenName( Scanner.Token ) );
  end;
Scanner.GetToken;
end; {GetAxis}

Function TBOSCompiler.GetFunc : string;
var
  Scope : TIdentScope;
  ScopeNameIsIdent : boolean;
begin
if Scanner.Token = CSTI_Identifier then
  begin
  Scope := CurrentIdentScope.FindSubScope( Scanner.Value, ScopeNameIsIdent );
  if (Scope = nil) then
    begin
    if ScopeNameIsIdent then
      Raise ECompileError.create( 'Function name "'+Scanner.Value+'" is used by a variable ' )
    else
      Raise ECompileError.create( 'Undeclaired function name, ' + Scanner.Value );
    end;  
  result := Scope.ScopeName;
  Scanner.GetToken;
  end
else
  begin
  result := Scanner.Value;
  WriteErrorMsg( 'Expecting function name, but found a ' + TokenName( Scanner.Token ) );
  end;
end; {GetAxis}

Function TBOSCompiler.GetLabel : integer;
begin
if CurrentIdentScope <> GlobalIdentScope then
  begin
  result := CurrentIdentScope.LabelCount;
  inc( CurrentIdentScope.LabelCount );
  end
else
  begin
  result := -1;
  WriteErrorMsg( 'Labels are not allowed in the global scope' );
  end;
end; {GetLabel}

Procedure TBOSCompiler.DoPush( value : TIdent );
begin
assert( value <> nil );
case value.Flags of
  IsLocalVar :
    begin
    WriteMsg( 'pushl ' + value.Name + ';');
    end;
  IsGlobalVar :
    begin
    WriteMsg( 'pushs ' + value.Name + ';' );
    end;
  IsPiece :
    begin
    WriteMsg( 'pushc ' + IntTostr( value.IndexValue ) + '; // '+ value.Name );
    end; // todo
  IsConst :
    begin
    WriteMsg( 'pushc ' + IntTostr( value.IndexValue ) + '; // '+ value.Name );
    end; // todo
  else
    begin
    WriteErrorMsg( 'Unknown result type' );
    WriteMsg( 'push ' + value.Name + ';' );
    end;
end;
end; {DoPush}

Procedure TBOSCompiler.DoPop( value : TIdent );
begin
assert( value <> nil );
case value.Flags of
  IsLocalVar :
    begin
    WriteMsg( 'popl ' + value.Name + ';');
    end;
  IsGlobalVar :
    begin
    WriteMsg( 'pops ' + value.Name + ';' );
    end;
  IsPiece,
  IsConst :
    begin
    WriteErrorMsg( 'Can not assign to a read-only identifier, ' + value.Name );
    WriteMsg( '// pops ' + value.Name + ';' );
    end;
  else
    begin
    WriteErrorMsg( 'Unknown result type, or result is readonly' );
    WriteMsg( 'pop ' + value.Name + ';' );
    end;
end;
end; {DoPop}

Procedure TBOSCompiler.DoPushConst( value : Longint );
begin
WriteMsg( 'pushc ' + IntToStr( value ) + ';' );
end; {DoPushConst}


Procedure TBOSCompiler.DoGlobalDeclarations;
begin
assert( self <> nil );
while true do
  case Scanner.Token of
    CSTII_piece : DoPiece;
    CSTII_GlobalVar : DoGlobalVar;
    CSTI_Identifier : DoFunction;
    CSTII_const : DoConst;
  else break;
  end;
end; {DoGlobalDeclarations}

Procedure TBOSCompiler.DoConst;
var
  NamedConstant : TIdent;
  ConstList : string;
begin
Scanner.Match( CSTII_const );
ConstList := 'const ';
try
  repeat
    if Scanner.Token <> CSTI_Identifier then
      Scanner.Match( CSTI_Identifier );
    NamedConstant := NewIdent( Scanner.Value, IsConst );
    assert( NamedConstant <> nil );
    ConstList := ConstList + Scanner.Value + ' ';
    Scanner.GetToken;
    Scanner.Match( CSTI_Assignment );
    ConstList := ConstList + '= ';
    NamedConstant.IndexValue := DoConstExpression( Scanner );
    ConstList := ConstList + IntToStr( NamedConstant.IndexValue) + ', ';
  until not Scanner.IfMatch( CSTI_Comma );
  setlength( ConstList, length( ConstList )-2 );
  Scanner.Match( CSTI_SemiColon );
  ConstList := ConstList +';';  
finally
  WriteStatusMsg( ConstList );
end;
end; {DoConst}

Procedure TBOSCompiler.DoPiece;
var
  PieceList : string;
begin
Scanner.Match( CSTII_piece );
PieceList := 'piece ';
try
  repeat
    if Scanner.Token <> CSTI_Identifier then
      Scanner.Match( CSTI_Identifier );
    NewIdent( Scanner.Value, IsPiece );
    PieceList := PieceList + Scanner.Value + ', ';
    Scanner.GetToken;
  until not Scanner.IfMatch( CSTI_Comma );
  setlength( PieceList, length( PieceList )-2 );
  Scanner.Match( CSTI_SemiColon );
  PieceList := PieceList + ';';
finally
  WriteMsg( PieceList );
end;
end; {DoPiece}

Procedure TBOSCompiler.DoGlobalVar;
var
  VarList : string;
begin
Scanner.Match( CSTII_GlobalVar );
VarList := 'static-var ';
try
  repeat
    if Scanner.Token <> CSTI_Identifier then
      Scanner.Match( CSTI_Identifier );
    NewIdent( Scanner.Value,IsGlobalVar );
    VarList := VarList + Scanner.Value + ', ';
    Scanner.GetToken;
  until not Scanner.IfMatch( CSTI_Comma );
  setlength( VarList, length( VarList )-2 );  
  Scanner.Match( CSTI_SemiColon );
  VarList := VarList + ';';
finally
  WriteMsg( VarList );
end;
end; {DoGlobalVar}

Procedure TBOSCompiler.DoFunction;
var
  FunctionName,ParamList : string;
  index,StartProcLineNumber : integer;
begin
if Scanner.Token <> CSTI_Identifier then
  Scanner.Match( CSTI_Identifier );
FunctionName := Scanner.Value;
Scanner.GetToken;
ActivateScope( FunctionName );
try try
  WriteMsg( '' );
  WriteMsg( 'proc ' + FunctionName);
  ParamList := '';
  // parse the param list
  Scanner.Match( CSTI_OpenRound );
  While Scanner.Token = CSTI_Identifier do
    begin
    ParamList := ParamList + Scanner.Value;
    NewIdent( Scanner.Value,IsLocalVar );
    Scanner.GetToken;
    if not Scanner.IfMatch( CSTI_Comma ) then
      break
    else
      ParamList := ParamList + ',';  
    end;
  Scanner.Match( CSTI_CloseRound );
  if ParamList = '' then
    ParamList := 'none';
  index := 1+WriteMsg( 'args ' + ParamList+';');
  StartProcLineNumber := WriteMsg( 'beginproc' );
  // do the actual function body
  LoopCount := 0;
  DoStatementBlock( true );
  // check to see if the proc is empty
  if (AsmListing.Count - StartProcLineNumber <= 3) then
    begin
    AsmListing.Delete(AsmListing.Count-1);
    AsmListing.Delete(AsmListing.Count-1);
    WriteMsg( 'empty;' );  
    end;
  if LocalVarList <> '' then
    setlength( LocalVarList, length( LocalVarList ) -2);
  if LocalVarList = '' then
    LocalVarList := 'none';
  WriteMsg( index, 'var ' + LocalVarList+';');
  LocalVarList := '';
  if (AnsiPos('ret',AsmListing[AsmListing.count-1]) <= 0) and
     AutoEmitLastRet then
    begin
    DoPushConst( 0 );
    WriteMsg( 'ret;' );
    end;
  WriteMsg( 'endproc;' );
  WriteMsg( '' );  
finally
  DeactivateCurrentScope;
end;
except
  On E : Exception do
  Raise ECompileError.Create( 'In function "' + FunctionName + '" ' + E.message ) at ExceptAddr;
end;
end; {DoFunction}

Procedure TBOSCompiler.DoStatementBlock( ForceBlock : boolean = false );
begin
inc( IndentLevel, 2 );
try
  If ForceBlock or Scanner.IfMatch( CSTII_Begin ) then
    begin
    if ForceBlock then
      Scanner.Match( CSTII_Begin );
    WriteMsg( '// {' );
    while DoStatement do ;
    Scanner.match( CSTII_End );
    WriteMsg( '// }' );
    end
  else
    DoStatement;
finally
  dec( IndentLevel, 2 );
end;
end; {DoStatementBlock}

function TBOSCompiler.DoStatement : boolean;
begin
result := true;
case Scanner.Token of
  CSTII_LocalVar : DoLocalVar;
  CSTII_const : DoConst;
  CSTII_if : DoIf;
  CSTII_While : DoWhile;
  CSTII_do : DoDoWhile;
  CSTII_for : DoFor;
  CSTII_Break : DoBreak;
  CSTII_Continue : DoContinue;
  CSTII_Return : DoReturn;
  CSTI_SemiColon : Scanner.Match( CSTI_SemiColon ); // null statement
  CSTI_Identifier: DoAssignment;
  CSTII_BASMInline : DoBASMInline;
//  CSTII_GetUnitProperty : DoGetUnitProperty;
//  CSTII_Rand : DoRand;
  CSTII_SetUnitProperty : DoSetUnitProperty;
  CSTII_Sleep : DoSleep;
  CSTII_Turn : DoTurn;
  CSTII_WaitForTurn : DoWaitForTurn;
  CSTII_Move : DoMove;
  CSTII_WaitForMove : DoWaitForMove;
  CSTII_Spin :  DoSpin;
  CSTII_StopSpin : DoStopSpin;
  CSTII_Hide : DoHide;
  CSTII_Show : DoShow;
  CSTII_Cache : DoCache;
  CSTII_DontCache : DoDontCache;
  CSTII_CallScript : DoCallScript( false, 0, false);
  CSTII_StartScript : DoStartScript( false, 0 );
  CSTII_Signal : DoSignal;
  CSTII_SetSignalMask : DoSetSignalMask;
  CSTII_Explode : DoExplode;
  CSTII_EmitSfx : DoEmitSfx;
  CSTII_AttachUnit : DoAttachUnit;
  CSTII_DropUnit : DoDropUnit;
  CSTII_Shade : DoShade;
  CSTII_DontShade : DoDontShade;
  CSTI_Increment : DoPreIncrement;
  CSTI_Decrement : DoPreDecrement;
else result := false;
end; {case}
end; {DoStatement}

Procedure TBOSCompiler.DoLocalVar;
var
  LocalVarIdent : TIdent;
  OutputLine : integer;
begin
Scanner.Match( CSTII_LocalVar );
repeat
  if Scanner.Token <> CSTI_Identifier then
    Scanner.Match( CSTI_Identifier );
  LocalVarIdent := NewIdent( Scanner.Value,IsLocalVar );
  LocalVarList := LocalVarList + Scanner.Value + ', ';
  Scanner.GetToken;
  if (Scanner.Token = CSTI_Assignment) then
    begin
    WriteBeginStatusMsg( OutputLine, '{MakeLocalVar }' + Scanner.Value );
    WritePartialStatusMsg( OutputLine, ' =');
    DoPush( LocalVarIdent );
    DoExpression( OutputLine );
    DoPop( LocalVarIdent );
    end;
until not Scanner.IfMatch( CSTI_Comma );
Scanner.Match( CSTI_SemiColon );
{
setlength( VarList, length( VarList )-2 );
WriteMsg( 'MakeLocalVar ' + VarList );
}
end; {DoLocalVar}

Procedure TBOSCompiler.DoIf;
var
  Label1 : integer;
  Label2 : integer;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'if ( ');
Label1 := GetLabel;
Scanner.Match( CSTII_if );
Scanner.Match( CSTI_OpenRound );
DoExpression( OutputLine );
WriteMsg( 'neg;' );
WriteMsg( 'jnz iflabel_' + IntToStr( Label1 ) + ';' );
Scanner.Match( CSTI_CloseRound );
WritePartialStatusMsg( OutputLine, ' )');
DoStatementBlock;
If Scanner.IfMatch( CSTII_Else ) then
  begin
  WriteBeginStatusMsg( OutputLine, 'else');
  Label2 := GetLabel;
  WriteMsg( 'jmp if_end_label_' + IntToStr( Label2 ) + ';' );
  WriteMsg( 'iflabel_' + IntToStr( Label1 ) + ':' );
  DoStatementBlock;
  WriteMsg( 'if_end_label_' + IntToStr( Label2 ) + ':' );
  end
else
  WriteMsg( 'iflabel_' + IntToStr( Label1 ) + ':' );
WriteMsg( '' );
end; {DoIf}

Procedure TBOSCompiler.DoWhile;
var
  Label1 : integer;
  Label2 : integer;
  OldLoopContinueLabel, OldLoopBreakLabel : string;
  OldUsedBreak : boolean;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'While (');
OldLoopContinueLabel := LoopContinueLabel;
OldLoopBreakLabel := LoopBreakLabel;
OldUsedBreak := UsedBreak;
inc( LoopCount );
try
  Label1 := GetLabel;
  Label2 := GetLabel;
  LoopContinueLabel := 'ConditionLabel_' + IntToStr( Label1 );
  LoopBreakLabel := 'TerminatorLabel_' + IntToStr( Label2 );
  WriteMsg( LoopContinueLabel + ':' );
  WritePartialStatusMsg( OutputLine, ' ) ');
  Scanner.Match( CSTII_while );
  Scanner.Match( CSTI_OpenRound );
  DoExpression( OutputLine );
  WriteMsg( 'jnz ' + LoopBreakLabel + ';' );
  Scanner.Match( CSTI_CloseRound );
  DoStatementBlock;
  WriteMsg( 'jmp ' + LoopContinueLabel + ';' );
  WriteMsg( LoopBreakLabel + ':' );
  WriteMsg( '' );
finally
  LoopContinueLabel := OldLoopContinueLabel;
  LoopBreakLabel := OldLoopBreakLabel;
  UsedBreak := OldUsedBreak;
  dec( LoopCount );
end;
end; {DoWhile}

Procedure TBOSCompiler.DoDoWhile;
var
  Label1 : integer;
  Label2 : integer;
  OldLoopContinueLabel, OldLoopBreakLabel : string;
  OldUsedBreak : boolean;
  OutputLine : integer;
begin
OldLoopContinueLabel := LoopContinueLabel;
OldLoopBreakLabel := LoopBreakLabel;
OldUsedBreak := UsedBreak;
inc( LoopCount );
try
  Label1 := GetLabel;
  Label2 := GetLabel;
  LoopContinueLabel := 'ConditionLabel_' + IntToStr( Label1 );
  LoopBreakLabel := 'TerminatorLabel_' + IntToStr( Label2 );
  WriteMsg( LoopContinueLabel + ':' );
  WriteBeginStatusMsg( OutputLine, 'Do');
  Scanner.Match( CSTII_Do );
  DoStatementBlock;
  WriteBeginStatusMsg( OutputLine, 'while (');
  Scanner.Match( CSTII_while );
  Scanner.Match( CSTI_OpenRound );
  DoExpression( OutputLine );
  WriteMsg( 'neg;' );
  WriteMsg( 'jnz ' + LoopContinueLabel + ';' );
  WritePartialStatusMsg( OutputLine, ' ) ');
  Scanner.Match( CSTI_CloseRound );
  if UsedBreak then
    WriteMsg( LoopBreakLabel + ':' );
  WriteMsg( '' );
finally
  LoopContinueLabel := OldLoopContinueLabel;
  LoopBreakLabel := OldLoopBreakLabel;
  UsedBreak := OldUsedBreak;
  dec( LoopCount );
end;
end; {DoDoWhile}

Procedure TBOSCompiler.DoFor;
var
  Label1 : integer;
  Label2 : integer;
  OldLoopContinueLabel, OldLoopBreakLabel : string;
  OldUsedBreak : boolean;
   OutputLine : integer;
begin
OldLoopContinueLabel := LoopContinueLabel;
OldLoopBreakLabel := LoopBreakLabel;
OldUsedBreak := UsedBreak;
inc( LoopCount );
try
  Label1 := GetLabel;
  Label2 := GetLabel;
  LoopContinueLabel := 'ConditionLabel_' + IntToStr( Label1 );
  LoopBreakLabel := 'TerminatorLabel_' + IntToStr( Label2 );
  WriteBeginStatusMsg( OutputLine, 'for ( ');
  Scanner.Match( CSTII_for );
  Scanner.Match( CSTI_OpenRound );
  // start expression
  repeat
    DoExpression( OutputLine );
    if scanner.token = CSTI_Comma then
      WritePartialStatusMsg( OutputLine, ', ');
  until not Scanner.IfMatch( CSTI_Comma );
  WritePartialStatusMsg( OutputLine, ' ; ');
  Scanner.Match( CSTI_SemiColon );
  WriteMsg( LoopContinueLabel + ':' );
  repeat
    DoExpression( OutputLine );
    if scanner.token = CSTI_Comma then
      WritePartialStatusMsg( OutputLine, ', ');
  until not Scanner.IfMatch( CSTI_Comma );
  WriteMsg( 'jnz ' + LoopBreakLabel + ';' );
  WritePartialStatusMsg( OutputLine, ' ; ');
  Scanner.Match( CSTI_SemiColon );
  repeat
    DoExpression( OutputLine );
    if scanner.token = CSTI_Comma then
      WritePartialStatusMsg( OutputLine, ', ');
  until not Scanner.IfMatch( CSTI_Comma );
  WritePartialStatusMsg( OutputLine, ' ) ');
  Scanner.Match( CSTI_CloseRound );
  DoStatementBlock;
  WriteMsg( 'jmp ' + LoopContinueLabel + ';' );
  WriteMsg( LoopBreakLabel + ':' );
  WriteMsg( '' );
finally
  LoopContinueLabel := OldLoopContinueLabel;
  LoopBreakLabel := OldLoopBreakLabel;
  UsedBreak := OldUsedBreak;  
  dec( LoopCount );
end;
end; {DoFor}

Procedure TBOSCompiler.DoBreak;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'Break' );
Scanner.Match( CSTII_Break );
//WriteDWord( 
//WriteMsg( 'break' );
WriteMsg( 'jmp ' + LoopBreakLabel + ';' );
UsedBreak := true;
end; {DoBreak}

Procedure TBOSCompiler.DoContinue;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'Continue' );
Scanner.Match( CSTII_Continue );
WriteMsg( 'jmp ' + LoopContinueLabel + ';' );
end; {DoContinue}

Procedure TBOSCompiler.DoReturn;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'return ' );
Scanner.Match( CSTII_Return );
DoExpression( OutputLine );
WriteMsg( 'ret;' );
end; {DoReturn}

Procedure TBOSCompiler.DoGetUnitProperty( OutputLine : integer );
begin
WritePartialStatusMsg( OutputLine, 'Get ' );
Scanner.Match( CSTII_GetUnitProperty );
DoExpression( OutputLine );
if ( Scanner.Token = CSTI_OpenRound ) then
  begin
  BiuldParamList( OutputLine );
  WriteMsg( 'get;' );
  end
else
  WriteMsg( 'getuv;' );
//WriteInstruction( Opcode_GetUnitValue );
// TODO 3 : TBOSCompiler.DoGetUnitProperty finish properly
end; {DoGetUnitProperty}

Procedure TBOSCompiler.DoSetUnitProperty;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'Set ' );
Scanner.Match( CSTII_SetUnitProperty );
DoExpression( OutputLine );
WriteBeginStatusMsg( OutputLine, 'to ' );
Scanner.Match( CSTII_To );
DoExpression( OutputLine );
WriteMsg( 'set;' );
end; {DoSetUnitProperty}

Procedure TBOSCompiler.DoSleep;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'sleep ' );
Scanner.Match( CSTII_Sleep ); 
DoExpression( OutputLine );
WriteMsg( 'sleep;' );
end; {DoSleep}

Procedure TBOSCompiler.DoTurn;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'turn ' );
Scanner.Match( CSTII_Turn );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' to ' );
Scanner.Match( CSTII_To );
GetAxis( aAxisName );
WritePartialStatusMsg( OutputLine, aAxisName+' ' );
DoExpression( OutputLine );
If scanner.IfMatch( CSTII_Speed )  then
  begin
  WritePartialStatusMsg( OutputLine, 'speed ' );
  DoExpression( OutputLine );
  end
else
  begin
  scanner.Match( CSTII_Now );
  WritePartialStatusMsg( OutputLine, 'now ' );
  end;
WriteMsg( 'turn;');
WriteMsg( 'ddw ' + aPieceName + ';' );
WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoTurn}

Procedure TBOSCompiler.DoWaitForTurn;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'wait-for-turn ' );
Scanner.Match( CSTII_WaitForTurn );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' around ' );
Scanner.Match( CSTII_around );
GetAxis( aAxisName );
WritePartialStatusMsg( OutputLine, aAxisName+' ' );
WriteMsg( 'wait-for-turn;');
WriteMsg( 'ddw ' + aPieceName + ';' );
WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoWaitForTurn}

Procedure TBOSCompiler.DoMove;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'move ' );
Scanner.Match( CSTII_Move );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' to ' );
Scanner.Match( CSTII_To );
GetAxis( aAxisName );
WritePartialStatusMsg( OutputLine, aAxisName+' ' );
DoExpression( OutputLine );
If not scanner.IfMatch( CSTII_Now ) then
  begin
  WritePartialStatusMsg( OutputLine, 'Speed ' );
  scanner.Match( CSTII_Speed );
  DoExpression( OutputLine );
  end
else
  begin
  WritePartialStatusMsg( OutputLine, 'Now ' );
  end;
WriteMsg( 'move; ' );
WriteMsg( 'ddw ' + aPieceName + ';' );
WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoMove}

Procedure TBOSCompiler.DoWaitForMove;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'wait-for-move ' );
Scanner.Match( CSTII_WaitForMove );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' along ' );
Scanner.Match( CSTII_along );
GetAxis( aAxisName );
WritePartialStatusMsg( OutputLine, aAxisName+' ' );
WriteMsg( 'wait-for-move;');
WriteMsg( 'ddw ' + aPieceName + ';' );
WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoWaitForMove}

Procedure TBOSCompiler.DoSpin;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin { TODO : lookup full syntax for spin }
WriteBeginStatusMsg( OutputLine, 'spin ' );
Scanner.Match( CSTII_Spin );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' around ' );
Scanner.Match( CSTII_around );
GetAxis( aAxisName );
WritePartialStatusMsg( OutputLine, aAxisName+' ' );
DoExpression( OutputLine );
WritePartialStatusMsg( OutputLine, 'speed ' );
scanner.Match( CSTII_Speed );
DoExpression( OutputLine );
If scanner.IfMatch( CSTII_Accelerate ) then
  begin
  WritePartialStatusMsg( OutputLine, 'accelerate ' );
  DoExpression( OutputLine );
  end
else
  DoPushConst( 0 );
WriteMsg( 'spin;');
WriteMsg( 'ddw ' + aPieceName + ';' );
WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoSpin}

Procedure TBOSCompiler.DoStopSpin;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin { TODO : lookup full syntax for stop-spin }
WriteBeginStatusMsg( OutputLine, 'stop-spin ' );
Scanner.Match( CSTII_StopSpin );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' around ' );
Scanner.Match( CSTII_around );
GetAxis( aAxisName );
WritePartialStatusMsg( OutputLine, aAxisName+' ' );
DoExpression( OutputLine );
If scanner.IfMatch( CSTII_Decelerate ) then
  begin
  WritePartialStatusMsg( OutputLine, 'decelerate ' );
  DoExpression( OutputLine );
  end
else
  DoPushConst( 0 );
WriteMsg( 'stop-spin;');
WriteMsg( 'ddw ' + aPieceName + ';' );
WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoStopSpin}

Procedure TBOSCompiler.DoHide;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'hide ' );
Scanner.Match( CSTII_Hide );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ' );
WriteMsg( 'hide;');
WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoHide}

Procedure TBOSCompiler.DoShow;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'show ' );
Scanner.Match( CSTII_Show );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ' );
WriteMsg( 'show;');
WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoShow}

Procedure TBOSCompiler.DoCache;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'cache  ' );
Scanner.Match( CSTII_Cache );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ' );
WriteMsg( 'cache;');
WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoCache}

Procedure TBOSCompiler.DoDontCache;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'dont-cache  ' );
Scanner.Match( CSTII_DontCache );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ' );
WriteMsg( 'dont-cache;');
WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoDontCache}

Procedure TBOSCompiler.DoShade;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'shade  ' );
Scanner.Match( CSTII_Shade );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ' );
WriteMsg( 'shade;');
WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoShade}

Procedure TBOSCompiler.DoDontShade;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'dont-shade  ' );
Scanner.Match( CSTII_DontShade );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ' );
WriteMsg( 'dont-shade;');
WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoDontShade}

Function TBOSCompiler.BiuldParamList( OutputLine : integer ) : integer;
begin
result := 0;
WritePartialStatusMsg( OutputLine, '( ');
Scanner.Match( CSTI_OpenRound );
While not Scanner.IfMatch( CSTI_CloseRound ) do
  begin
  DoExpression( OutputLine );
  inc( result );
  if not Scanner.IfMatch( CSTI_Comma ) then
    begin
    Scanner.Match( CSTI_CloseRound );
    break;
    end;
  WritePartialStatusMsg( OutputLine, ', ');
  end;
WritePartialStatusMsg( OutputLine, ' )');
end; {BiuldParamList}

Procedure TBOSCompiler.DoCallScript( CanReturnValue : boolean{ = false};
                                     OutputLine : integer{ = 0};
                                     NoCallScriptMarker : boolean{ = false});
var
  FunctionName : string;
  ParamCount : integer;
begin
// DONE 2 : Allow call-script to return a value
if CanReturnValue then
  WritePartialStatusMsg( OutputLine, 'call-script ' )
else
  WriteBeginStatusMsg( OutputLine, 'call-script ' );
if not NoCallScriptMarker then
  Scanner.Match( CSTII_CallScript );
FunctionName := GetFunc;
WritePartialStatusMsg( OutputLine, FunctionName+' ' );
ParamCount := BiuldParamList( OutputLine );
//WriteInstruction( Opcode_CallScript );
//WriteInstrArg( FuncIndex );
//WriteInstrArg( ParamCount );
WriteMsg( 'call;');
WriteMsg( 'ddw ' + FunctionName + ';' );
WriteMsg( 'ddw ' + IntToStr( ParamCount ) + ';' );
end; {DoCallScript}

Procedure TBOSCompiler.DoStartScript( CanReturnValue : boolean{ = false};
                                      OutputLine : integer{ = 0});
var
  FunctionName : string;
  ParamCount : integer;
begin
// TODO 1 : Determine if "Start-Script"  can return anything meaning full
if CanReturnValue then
  WritePartialStatusMsg( OutputLine, 'start-script ' )
else
  WriteBeginStatusMsg( OutputLine, 'start-script ' );
Scanner.Match( CSTII_StartScript );
FunctionName := GetFunc;
WritePartialStatusMsg( OutputLine, FunctionName+' ' );
ParamCount := BiuldParamList( OutputLine );
//WriteInstruction( Opcode_StartScript );
//WriteInstrArg( FuncIndex );
//WriteInstrArg( ParamCount );
WriteMsg( 'start;');
WriteMsg( 'ddw ' + FunctionName + ';' );
WriteMsg( 'ddw ' + IntToStr( ParamCount ) + ';' );
end; {DoStartScript}

Procedure TBOSCompiler.DoSignal;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'signal ' );
Scanner.Match( CSTII_Signal );
DoExpression( OutputLine );
WriteMsg( 'sig;' );
end; {DoSignal}

Procedure TBOSCompiler.DoSetSignalMask;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'set-signal-mask ' );
Scanner.Match( CSTII_SetSignalMask );
DoExpression( OutputLine );
WriteMsg( 'mask;' );
end; {DoSetSignalMask}

Procedure TBOSCompiler.DoExplode;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'explode ' );
Scanner.Match( CSTII_Explode );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' type ');
Scanner.Match( CSTII_ExplodeType );
DoExpression( OutputLine );
WriteMsg( 'explode;' );
WriteMsg( 'ddw ' + aPieceName + ';');
end; {DoExplode}

Procedure TBOSCompiler.DoEmitSfx;
var
  aPieceName : string;
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'emit-sfx ' );
Scanner.Match( CSTII_EmitSfx );
DoExpression( OutputLine );
WritePartialStatusMsg( OutputLine, 'from ');
Scanner.Match( CSTII_from );
GetPiece( aPieceName );
WritePartialStatusMsg( OutputLine, aPieceName+' ');
WriteMsg( 'sfx; ' );
WriteMsg( 'ddw ' + aPieceName + ';');
end; {DoEmitSfx}

Procedure TBOSCompiler.DoRand( OutputLine : integer );
begin
WritePartialStatusMsg( OutputLine, 'rand( ' );
Scanner.Match( CSTII_Rand );
Scanner.Match( CSTI_OpenRound );
DoExpression( OutputLine );
WritePartialStatusMsg( OutputLine, ', ');
Scanner.Match( CSTI_Comma );
DoExpression( OutputLine );
WritePartialStatusMsg( OutputLine, ') ');
Scanner.Match( CSTI_CloseRound );
WriteMsg( 'rand;' );
end; {DoRand}

Procedure TBOSCompiler.DoAttachUnit;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'attach-unit ' );
Scanner.Match( CSTII_AttachUnit );
DoExpression( OutputLine );
WritePartialStatusMsg( OutputLine, 'to ' );
Scanner.Match( CSTII_to );
DoExpression( OutputLine );
DoPushConst( 0 );
WriteMsg( 'attach;' );
end; {DoAttachUnit}

Procedure TBOSCompiler.DoDropUnit;
var
  OutputLine : integer;
begin
WriteBeginStatusMsg( OutputLine, 'drop-unit ' );
Scanner.Match( CSTII_DropUnit );
DoExpression( OutputLine );
WriteMsg( 'drop;' );
end; {DoDropUnit}

Procedure TBOSCompiler.DoBASMInline;
var
  OutputLine : integer;
  BASMLine : string;
begin // DONE 2 : Allow inlining BASM code
WriteBeginStatusMsg( OutputLine, 'basm' );
Scanner.Match( CSTII_BASMInline );
inc( IndentLevel, 2 );
WriteBeginStatusMsg( OutputLine, '{' );
try
  Scanner.Match( CSTII_Begin );
  While not Scanner.IfMatch( CSTII_end ) do
    begin
    BASMLine := '';
    While not Scanner.IfMatch( CSTI_SemiColon ) and
          not Scanner.IfMatch( CSTII_end ) do
      begin
      BASMLine := BASMLine +Scanner.Value + ' ';
      Scanner.GetToken;
      end;
    Delete(BASMLine, length(BASMLine),1);
    BASMLine := BASMLine +';';
    WriteMsg( BASMLine );
    end;
finally
  WriteBeginStatusMsg( OutputLine, '}' );
  dec( IndentLevel, 2 );
end;
end; {DoBASMInline}

Procedure TBOSCompiler.AssignmentHelper( OutputLine : integer;
                                         var ResultIdent : TIdent);
var
  ResultName : string;
  a : boolean;                                           
begin
if Scanner.Token <> CSTI_Identifier then
  Scanner.Match( CSTI_Identifier );
ResultName := Scanner.Value;
WritePartialStatusMsg( OutputLine, ResultName +' ');
Scanner.GetToken;
ResultIdent := FindIdent( ResultName, a );
if ResultIdent = nil then
  begin
  WriteErrorMsg( 'Undeclaired identifier, ' + ResultName );
  ResultIdent := NewIdent( ResultName, UnknownVarType );
  end;
end; {AssignmentHelper}

Procedure TBOSCompiler.DoAssignment;
var
  ResultIdent : TIdent;
  OutputLine : integer;
//  OutputOpcode : longword;
  AssignmentType : TToken;
  ValidAssignmentTokens : set of TToken;
begin
WriteBeginStatusMsg( OutputLine, '');
AssignmentHelper( OutputLine, ResultIdent );
ValidAssignmentTokens := [ CSTI_Assignment_Plus, CSTI_Assignment_Minus,
                           CSTI_Assignment_Divide, CSTI_Assignment_Multiply,
                           CSTI_Assignment_BitwiseOr, CSTI_Assignment_BitwiseXor,
                           CSTI_Assignment_BitwiseAnd, CSTI_Assignment_LogicalOr,
                           CSTI_Assignment_LogicalXor, CSTI_Assignment_LogicalAnd ];
if EnableBASMExtentions then
  begin
  include( ValidAssignmentTokens, CSTI_Assignment_ShiftLeft );
  include( ValidAssignmentTokens, CSTI_Assignment_ShiftRight );
  end;
if Scanner.Token = CSTI_Assignment then
  begin
  WritePartialStatusMsg(OutputLine, '= ');
  Scanner.GetToken;
  DoExpression( OutputLine );
  end
else if Scanner.Token in ValidAssignmentTokens then
  begin
  AssignmentType := Scanner.Token;
  Scanner.GetToken;
  case AssignmentType of
    CSTI_Assignment_Plus       : WritePartialStatusMsg( '+= ' );
    CSTI_Assignment_Minus      : WritePartialStatusMsg( '-= ' );
    CSTI_Assignment_Divide     : WritePartialStatusMsg( '/= ' );
    CSTI_Assignment_Multiply   : WritePartialStatusMsg( '*= ' );
    CSTI_Assignment_BitwiseOr  : WritePartialStatusMsg( '|= ' );
    CSTI_Assignment_BitwiseXor : WritePartialStatusMsg( '^= ' );
    CSTI_Assignment_BitwiseAnd : WritePartialStatusMsg( '&= ' );
    CSTI_Assignment_LogicalOr  : WritePartialStatusMsg( '||= ' );
    CSTI_Assignment_LogicalXor : WritePartialStatusMsg( '^^= ' );
    CSTI_Assignment_LogicalAnd : WritePartialStatusMsg( '&&= ' );
    CSTI_Assignment_ShiftLeft  : WritePartialStatusMsg( '<<= ' );
    CSTI_Assignment_ShiftRight : WritePartialStatusMsg( '>>= ' );
  end; {case}
  DoPush( ResultIdent );
  DoExpression( OutputLine );
  case AssignmentType of
    CSTI_Assignment_Plus       : WriteMsg( 'add;' );
    CSTI_Assignment_Minus      : WriteMsg( 'sub;' );
    CSTI_Assignment_Divide     : WriteMsg( 'div;' );
    CSTI_Assignment_Multiply   : WriteMsg( 'mul;' );
    CSTI_Assignment_BitwiseOr  : WriteMsg( 'or;' );
    CSTI_Assignment_BitwiseXor : WriteMsg( 'xor;' );
    CSTI_Assignment_BitwiseAnd : WriteMsg( 'and;' );
    CSTI_Assignment_LogicalOr  : WriteMsg( 'lor;' );
    CSTI_Assignment_LogicalXor : WriteMsg( 'lxor;' );
    CSTI_Assignment_LogicalAnd : WriteMsg( 'land;' );
    CSTI_Assignment_ShiftLeft  : WriteMsg( 'shl;' );
    CSTI_Assignment_ShiftRight : WriteMsg( 'sar;' );
  end; {case}
  end
else
  begin
  WriteErrorMsg( 'Expecting assignment type' );
  Scanner.Match( CSTI_Assignment );
  end;
DoPop( ResultIdent );
//WriteInstruction( OutputOpcode );
//if OutputOpcode <> 0 then
//  WriteInstrArg( ResultIdent.IndexValue );
end; {DoAssignment}


Procedure TBOSCompiler.DoPreIncrement;
var
  OutputLine : longint;
  ResultIdent : TIdent;
begin
WriteBeginStatusMsg( OutputLine, '++ ' );
Scanner.Match( CSTI_Increment );
AssignmentHelper( OutputLine, ResultIdent );
DoPush( ResultIdent );
DoPushConst( 1 );
WriteMsg( 'add;' );
DoPop( ResultIdent );
end; {DoPreIncrement}

Procedure TBOSCompiler.DoPreDecrement;
var
  OutputLine : longint;
  ResultIdent : TIdent;
begin
WriteBeginStatusMsg( OutputLine, '-- ' );
Scanner.Match( CSTI_Decrement );
AssignmentHelper( OutputLine, ResultIdent );
DoPush( ResultIdent );
DoPushConst( 1 );
WriteMsg( 'sub;' );
DoPop( ResultIdent );
end; {DoPreDecrement}

// Parse and Translate a Math Factor
Procedure TBOSCompiler.DoFactor( OutputLine : integer );
var
  Ident : TIdent;
  IntValue : Int64;
  IdentNameIsScope : boolean;
begin
case Scanner.token of
  CSTII_GetUnitProperty : DoGetUnitProperty( OutputLine );
  CSTII_Rand : DoRand( OutputLine );
  CSTII_CallScript : DoCallScript( true, OutputLine, false );
  CSTI_Increment :
    begin
    WritePartialStatusMsg( OutputLine, '++ ' );
    Scanner.Match( CSTI_Increment );
    AssignmentHelper( OutputLine, ident );
    DoPush( ident );
    DoPushConst( 1 );
    WriteMsg( 'add;' );
    DoPop( ident );
    DoPush( ident );
    end;
  CSTI_Decrement :
    begin
    WriteBeginStatusMsg( OutputLine, '-- ' );
    Scanner.Match( CSTI_Decrement );
    AssignmentHelper( OutputLine, Ident );
    DoPush( ident );
    DoPushConst( 1 );
    WriteMsg( 'sub;' );
    DoPop( ident );
    DoPush( ident );
    end;
//  CSTII_StartScript : DoStartScript( true, OutputLine );  
  CSTI_OpenRound :
    begin
    WritePartialStatusMsg(OutputLine, '( ');
    Scanner.Match( CSTI_OpenRound );
    DoExpression( OutputLine );
    WritePartialStatusMsg(OutputLine, ') ');    
    Scanner.Match( CSTI_CloseRound );
    end;
  CSTI_Identifier :
    begin
    WritePartialStatusMsg(OutputLine, Scanner.Value + ' ');
    Ident := FindIdent( Scanner.Value, IdentNameIsScope );
    if Ident = nil then
       begin
       if not IdentNameIsScope then
         begin // referencing a non-existant variable
         WriteErrorMsg( 'Undeclaired identifier, ' + Scanner.Value );
         Ident := NewIdent( Scanner.Value, UnknownVarType );
         end
       else  // inlined function call
         DoCallScript( true, OutputLine, true );
       end;
    if not IdentNameIsScope then
      begin
      DoPush( Ident );
      Scanner.GetToken;
      case Scanner.token of
        CSTI_Increment :
          begin
          WritePartialStatusMsg( OutputLine, '++ ' );
          Scanner.Match( CSTI_Increment );
          DoPush( Ident );
          DoPushConst( 1 );
          WriteMsg( 'add;' );
          DoPop( Ident );
          end;
        CSTI_Decrement :
          begin
          WriteBeginStatusMsg( OutputLine, '-- ' );
          Scanner.Match( CSTI_Decrement );
          DoPush( Ident );
          DoPushConst( 1 );
          WriteMsg( 'sub;' );
          DoPop( Ident );
          end;
      end;    
      end;
    end;
  CSTI_Integer, CSTI_HexInt :
    begin
    WritePartialStatusMsg(OutputLine, Scanner.Value + ' ');
    IntValue := StrToInt64( Scanner.Value );
    if IntValue > High(Longword) then
      IntValue := IntValue mod High(Longword);
    DoPushConst( IntValue );
    Scanner.GetToken;
    end;
  CSTI_Real :
    begin
    WritePartialStatusMsg(OutputLine, Scanner.Value + ' ');
    WriteErrorMsg( 'Floatpoint values are not permitted in arbitrary locations ' + Scanner.Value );
    WriteMsg( 'pushc ' + Scanner.Value + ';' );
    Scanner.GetToken;
    end;
end;    
end; {DoFactor}

Procedure TBOSCompiler.DoUnaryFactor( OutputLine : integer );
var
  Token : TToken;
begin
while true do
  begin
  Token := Scanner.Token;
  if Token in [ CSTI_BitwiseNot, CSTI_LogicalNot, CSTI_Minus ] then
    begin
    WritePartialStatusMsg(OutputLine, TokenName( Scanner.Token )+ ' ');
    Scanner.GetToken;
    end;
  DoFactor( OutputLine );
  case Token of
    CSTI_BitwiseNot :
      begin
      WriteMsg( 'not;' );
      end;
    CSTI_LogicalNot :
      begin
      WriteMsg( 'neg;' );
      end;
    CSTI_Minus :
      begin
      DoPushConst( 0 );
      WriteMsg( 'sub;' );
      end;
  else break;
  end;
  end;
end; {DoUnaryFactor}

Procedure TBOSCompiler.DoMulTerm( OutputLine : integer );
begin
DoUnaryFactor( OutputLine );
while true do
case Scanner.Token of
  CSTI_Multiply :
    begin
    WritePartialStatusMsg(OutputLine, '* ');
    Scanner.Match( CSTI_Multiply );
    DoUnaryFactor( OutputLine );
    WriteMsg( 'mul;' );
    end;
  CSTI_Divide :
    begin
    WritePartialStatusMsg(OutputLine, '\ ');
    Scanner.Match( CSTI_Divide );
    DoUnaryFactor( OutputLine );
    WriteMsg( 'div;' );
    end;
  else break;
end;
end; {DoMulTerm}

Procedure TBOSCompiler.DoAddTerm( OutputLine : integer );
begin
DoMulTerm( OutputLine );
while true do
case Scanner.Token of
  CSTI_Plus :
    begin
    WritePartialStatusMsg(OutputLine, '+ ');
    Scanner.Match( CSTI_Plus );
    DoMulTerm( OutputLine );
    WriteMsg( 'add;' );
    end;
  CSTI_Minus :
    begin
    WritePartialStatusMsg(OutputLine, '- ');
    Scanner.Match( CSTI_Minus );
    DoMulTerm( OutputLine );
    WriteMsg( 'sub;' );
    end;
  else break;
end;
end; {DoAddTerm}

Procedure TBOSCompiler.DoShiftTerm( OutputLine : integer );
begin
DoAddTerm( OutputLine );
if EnableBASMExtentions then
while true do
case Scanner.Token of
  CSTI_ShiftLeft :
    begin
    WritePartialStatusMsg(OutputLine, '<< ');
    Scanner.Match( CSTI_ShiftLeft );
    DoAddTerm( OutputLine );
    WriteMsg( 'shl;' );
    end;
  CSTI_ShiftRight :
    begin
    WritePartialStatusMsg(OutputLine, '>> ');
    Scanner.Match( CSTI_ShiftRight );
    DoAddTerm( OutputLine );
    WriteMsg( 'sar;' );
    end;
  else break;
end;
end; {DoShiftTerm}

Procedure TBOSCompiler.DoRelationTerm( OutputLine : integer );
begin
DoShiftTerm( OutputLine );
case Scanner.Token of
  CSTI_Greater :
    begin
    WritePartialStatusMsg(OutputLine, '> ');
    Scanner.Match( CSTI_Greater );
    DoShiftTerm( OutputLine );
    WriteMsg( 'setg;' );
    end;
  CSTI_GreaterEqual :
    begin
    WritePartialStatusMsg(OutputLine, '>= ');
    Scanner.Match( CSTI_GreaterEqual );
    DoShiftTerm( OutputLine );
    WriteMsg( 'setge;' );
    end;
  CSTI_Less :
    begin
    WritePartialStatusMsg(OutputLine, '< ');
    Scanner.Match( CSTI_Less );
    DoShiftTerm( OutputLine );
    WriteMsg( 'setl;' );
    end;
  CSTI_LessEqual :
    begin
    WritePartialStatusMsg(OutputLine, '<= ');
    Scanner.Match( CSTI_LessEqual );
    DoShiftTerm( OutputLine );
    WriteMsg( 'setle;' );
    end;
end;
end; {DoRelationTerm}

Procedure TBOSCompiler.DoEqualityTerm( OutputLine : integer );
begin
DoRelationTerm( OutputLine );
case Scanner.Token of
  CSTI_Equal :
    begin
    WritePartialStatusMsg(OutputLine, '== ');
    Scanner.Match( CSTI_Equal );
    DoRelationTerm( OutputLine );
    WriteMsg( 'cmp;' );
    end;
  CSTI_NotEqual :
    begin
    WritePartialStatusMsg(OutputLine, '!= ');
    Scanner.Match( CSTI_NotEqual );
    DoRelationTerm( OutputLine );
    WriteMsg( 'setne;' );
    end;
end;
end; {DoEqualityTerm}

Procedure TBOSCompiler.DoExpression( OutputLine : integer );
begin
DoEqualityTerm( OutputLine );
while true do
  begin
  if Scanner.token = CSTI_EOF then
    break;
  if Scanner.IfMatch( CSTI_LogicalOr ) then
    begin
    WritePartialStatusMsg(OutputLine, '!| ');
    DoEqualityTerm( OutputLine );
    WriteMsg( 'lor;' );
    end
  else if Scanner.IfMatch( CSTI_LogicalXor ) then
    begin
    WritePartialStatusMsg(OutputLine, '^^ ');
    DoEqualityTerm( OutputLine );
    WriteMsg( 'lxor;' );
    end
  else if Scanner.IfMatch( CSTI_LogicalAnd ) then
    begin
    WritePartialStatusMsg(OutputLine, '&& ');
    DoEqualityTerm( OutputLine );
    WriteMsg( 'land;' );
    end
  else if Scanner.IfMatch( CSTI_BitwiseOr ) then
    begin
    WritePartialStatusMsg(OutputLine, '| ');
    DoEqualityTerm( OutputLine );
    WriteMsg( 'or;' );
    end
  else if Scanner.IfMatch( CSTI_BitwiseXor ) then
    begin
    WritePartialStatusMsg(OutputLine, '^ ');
    DoEqualityTerm( OutputLine );
    WriteMsg( 'xor;' );
    end
  else if Scanner.IfMatch( CSTI_BitwiseAnd ) then
    begin
    WritePartialStatusMsg(OutputLine, '& ');
    DoEqualityTerm( OutputLine );
    WriteMsg( 'and' );
    end
  else break;
  end;
end; {DoExpression}

end.

