unit ScannerU;

// DONE  : Implemented the operations "++i", "i++", "--i", "i--"

{$DEFINE ConvertCommentsToWhiteSpace}
{.$DEFINE BreakOnError}
{$DEFINE CaseSensitive}

interface
uses
  sysutils,
  classes;

const
  TAB = #09;
  LF = #10;
  CR = #13;
Type
  EScannerError = class(Exception);

  TToken = (
  {Items that are used internally}
    CSTINT_WhiteSpace,
  {Tokens}
    CSTI_EOF,
    CSTI_Identifier,
    CSTI_SemiColon,
    CSTI_Comma,
    CSTI_Period,
    CSTI_Colon,
    CSTI_OpenRound,
    CSTI_CloseRound,
    CSTI_OpenBlock,
    CSTI_CloseBlock,
    CSTI_Assignment,
    CSTI_Equal,
    CSTI_NotEqual,
    CSTI_Greater,
    CSTI_GreaterEqual,
    CSTI_Less,
    CSTI_LessEqual,
    CSTI_Plus,
    CSTI_Minus,
    CSTI_Divide,
    CSTI_Multiply,

    CSTI_Increment,
    CSTI_Decrement,

    CSTI_LogicalOr,
    CSTI_LogicalXor,
    CSTI_LogicalAnd,
    CSTI_LogicalNot,

    CSTI_BitwiseOr,
    CSTI_BitwiseXor,
    CSTI_BitwiseAnd,
    CSTI_BitwiseNot,

    CSTI_ShiftLeft,
    CSTI_ShiftRight,

    CSTI_Assignment_Plus,
    CSTI_Assignment_Minus,
    CSTI_Assignment_Divide,
    CSTI_Assignment_Multiply,

    CSTI_Assignment_LogicalOr,
    CSTI_Assignment_LogicalXor,
    CSTI_Assignment_LogicalAnd,

    CSTI_Assignment_BitwiseOr,
    CSTI_Assignment_BitwiseXor,
    CSTI_Assignment_BitwiseAnd,

    CSTI_Assignment_ShiftLeft,
    CSTI_Assignment_ShiftRight,

    CSTI_Integer,
    CSTI_HexInt,
    CSTI_Real,
    CSTI_String,

  {Identifiers}
    CSTII_Begin,
    CSTII_End,
    
    CSTII_GlobalVar,
    CSTII_LocalVar,
    CSTII_piece,
    CSTII_const,

    CSTII_BASMInline,

    CSTII_GetUnitProperty,
    CSTII_SetUnitProperty,
    CSTII_Sleep,
    CSTII_WaitForTurn,
    CSTII_WaitForMove,
    CSTII_Turn,
    CSTII_Move,
    CSTII_Spin,
    CSTII_StopSpin,
    CSTII_To,
    CSTII_Around,
    CSTII_along,
    CSTII_from,
    CSTII_Now,
    CSTII_Speed,
    CSTII_Accelerate,
    CSTII_Decelerate,
    CSTII_Hide,
    CSTII_Show,
    CSTII_Cache,
    CSTII_DontCache,
    CSTII_CallScript,
    CSTII_StartScript,
    CSTII_Signal,
    CSTII_SetSignalMask,
    CSTII_Explode,
    CSTII_ExplodeType,
    CSTII_EmitSfx,
    CSTII_Shade,
    CSTII_DontShade,
    CSTII_rand,
    CSTII_AttachUnit,
    CSTII_DropUnit,

    CSTII_for,
    CSTII_if,
    CSTII_else,
    CSTII_do,
    CSTII_while,
    CSTII_break,
    CSTII_Continue,
    CSTII_return
    );
    
  TTokens = set of TToken;

  TTokenData = record
    Token : TToken;
    value : string;
    Position : Integer;
  end; {TTokenData}

  TScanner = class;

  TScanner = Class
  Private
    ftext : string;
    // the index into the source text of the next char
    fNextCharIndex : integer;
    // look ahead char
    Look,NextLook : char;
    fToken : TTokenData;
    fNextToken : TTokenData;
    // the key word table
    KeyWords : TStringList;
   {
    Table Lookup
    If the input string matches a table entry, return the entry
    index.  If not, return a zero. (case-sensitive)
   }
    function KeyWordLookup(Const Symbol : string): TToken;

    // error methods
    procedure Error(Const s: string; ErrorAddress: Pointer);

    procedure Expected(Const s,a: string); overload;
    procedure Expected(Const s,a: string; ErrorAddress: Pointer); overload;
    Procedure Scanner_Abort(Const s : string);

    // Read New Character From Input Stream
    function GetChar : char;
    // Skip Over a commented section, if 'ConvertCommentsToWhiteSpace' is defined,
    // the comments are converted into a white space char
    Procedure SkipComment;
    // Skip Over Leading White Space
    procedure SkipWhite;

    // Get an Identifier
    Procedure GetIdentifier;
    // Get a string
    procedure GetString;    
    // Get a Number
    Procedure GetNumber;
    Procedure GetOperator;

  Public
    Constructor Create( Const atext : string);
    Destructor Destroy; override;

    // Skip Over a Comma if it exists.
    function SkipComma : boolean;

    // loads a token into 'token', and its value into 'value'
    // the value reported is a step behind the scanner, this allows
    // the next token (the current token for the scanner) to be access able to the
    // Parser
    Procedure GetToken;  // Lexical Scanner

    // Skips the the next new line
    procedure SkipToNewLine;
    // returns the string of characters before the next newline
    Function TillNewLine : string;
    // Inserts a character into the text stream
    Procedure InsertString( index : integer;
                            const substr : string;
                            AutoSkipWhiteSpace: boolean = true );
    Procedure RemoveString( index, count : integer);

    Procedure Match( aToken : TToken );
    Function IfMatch( aToken : TToken ): boolean;

    Property CurrentToken : TTokenData read fToken write fToken;
    Property NextToken : TTokenData read fnextToken;

    Property Token : TToken read fToken.Token write fToken.Token;
    Property Value : string read fToken.Value write fToken.Value;
    Property Position : integer read fToken.Position write fToken.Position;

    Property NextCharIndex : integer read fNextCharIndex write fNextCharIndex;
    Property Text : string read ftext;
  end; {TScanner}

Const
  StringChar = '"';
  PreProcessorCommand = '#';
  HexadecimalChar = '$';

  CommentType1      = '//';  // terminates at a new line
  CommentType2Start = '/*';  // start Multi-line comment
  CommentType2End   = '*/';  // end Multi-line comment

// Definition of Keywords and Token Types
type
  TSymbolData = record
   Symbol : string;
   Token : TToken;
  end; {TSymbolData}
const
  KeyWordCout = 46;
  KeyWordlist: array [0..KeyWordCout-1] of TSymbolData = (
      (Symbol : 'const'; Token : CSTII_const),
      (Symbol : 'return'; Token : CSTII_return),
      (Symbol : 'break'; Token : CSTII_break),
      (Symbol : 'continue'; Token : CSTII_Continue),
      (Symbol : 'for'; Token : CSTII_for),
      (Symbol : 'if'; Token : CSTII_if),
      (Symbol : 'else'; Token : CSTII_else),
      (Symbol : 'do'; Token : CSTII_do),
      (Symbol : 'while'; Token : CSTII_while),
      (Symbol : 'static-var'; Token : CSTII_GlobalVar),
      (Symbol : 'var'; Token : CSTII_LocalVar),
      (Symbol : 'piece'; Token : CSTII_piece),
      (Symbol : 'basm'; Token : CSTII_BASMInline),

{
      (Symbol : 'or'; Token : CSTI_LogicalOr),
      (Symbol : 'xor'; Token : CSTI_LogicalXor),
      (Symbol : 'and'; Token : CSTI_LogicalAnd),
      (Symbol : 'not'; Token : CSTI_LogicalNot),
}

      (Symbol : 'get'; Token : CSTII_GetUnitProperty),
      (Symbol : 'set'; Token : CSTII_SetUnitProperty),
      (Symbol : 'sleep'; Token : CSTII_Sleep),
      (Symbol : 'wait-for-turn'; Token : CSTII_WaitForTurn),
      (Symbol : 'wait-for-move'; Token : CSTII_WaitForMove),
      (Symbol : 'turn'; Token : CSTII_Turn),
      (Symbol : 'move'; Token : CSTII_Move),
      (Symbol : 'spin'; Token : CSTII_Spin),
      (Symbol : 'stop-spin'; Token : CSTII_StopSpin),
      (Symbol : 'to'; Token : CSTII_To),
      (Symbol : 'around'; Token : CSTII_Around),
      (Symbol : 'along'; Token : CSTII_along),
      (Symbol : 'from'; Token : CSTII_from),
      (Symbol : 'now'; Token : CSTII_Now),
      (Symbol : 'speed'; Token : CSTII_Speed),
      (Symbol : 'accelerate'; Token : CSTII_Accelerate),
      (Symbol : 'decelerate'; Token : CSTII_Decelerate),
      (Symbol : 'hide'; Token : CSTII_Hide),
      (Symbol : 'show'; Token : CSTII_Show),
      (Symbol : 'cache'; Token : CSTII_Cache),
      (Symbol : 'dont-cache'; Token : CSTII_DontCache),
      (Symbol : 'call-script'; Token : CSTII_CallScript),
      (Symbol : 'start-script'; Token : CSTII_StartScript),
      (Symbol : 'signal'; Token : CSTII_Signal),
      (Symbol : 'set-signal-mask'; Token : CSTII_SetSignalMask),
      (Symbol : 'explode'; Token : CSTII_Explode),
      (Symbol : 'type'; Token : CSTII_ExplodeType),
      (Symbol : 'emit-sfx'; Token : CSTII_EmitSfx),
      (Symbol : 'shade'; Token : CSTII_Shade),
      (Symbol : 'dont-shade'; Token : CSTII_DontShade),
      (Symbol : 'rand'; Token : CSTII_rand),
      (Symbol : 'attach-unit'; Token : CSTII_AttachUnit),
      (Symbol : 'drop-unit'; Token : CSTII_DropUnit)
      );

const
  AngularConstant = 65536 / 360;
  LinearConstant = 163840.0;
var
  AngularConst : Extended = AngularConstant;
  LinearConst : Extended  = LinearConstant;

function MakeKeywordTable : TStringlist;

Function TokenName( aToken : TToken) : string;


// Recognize a Decimal Digit
function IsDigit(c: char): boolean;
// Recognize an Alpha Character
function IsAlpha(c: char): boolean;
// Recognize an AlphaNumeric Character
function IsAlNum(c: char): boolean;
// Recognize newline Characters
function IsNewLine(c: char) : boolean;
// Recognize Any Operator
function IsOp(c: char): boolean;

implementation

Function TokenName( aToken : TToken) : string;
begin
case aToken of
  CSTINT_WhiteSpace: Result := 'whitespace (internal, should never be seen)';
{Tokens}
  CSTI_EOF: Result := 'End of File';
  CSTI_Identifier: Result := 'identifier';
  CSTI_SemiColon: Result := ';';
  CSTI_Comma: Result := ':';
  CSTI_Period: Result := '.';
  CSTI_Colon: Result := ',';
  CSTI_OpenRound: Result := '(';
  CSTI_CloseRound: Result := ')';
  CSTI_OpenBlock: Result := '[';
  CSTI_CloseBlock: Result := ']';

  CSTI_Assignment: Result := '=';
  CSTI_Divide: Result := '/';
  CSTI_Multiply: Result := '*';
  CSTI_Plus: Result := '+';
  CSTI_Minus: Result := '-';
  CSTI_Increment: Result := '++';
  CSTI_Decrement: Result := '--';
  CSTI_Greater: Result := '>';
  CSTI_GreaterEqual: Result := '>=';
  CSTI_Less: Result := '<';
  CSTI_LessEqual: Result := '<=';
  CSTI_Equal: Result := '==';
  CSTI_NotEqual: Result := '!=';
  CSTI_LogicalOr: Result := '|| (logical or)';
  CSTI_LogicalXor: Result := '^^ (logical xor)';
  CSTI_LogicalAnd: Result := '&& (logical and)';
  CSTI_LogicalNot: Result := '~ (logical not)';
  CSTI_BitwiseOr: Result := '| (bitwise or)';
  CSTI_BitwiseXor: Result := '^ (bitwise xor)';
  CSTI_BitwiseAnd: Result := '& (bitwise and)';
  CSTI_BitwiseNot: Result := '! (bitwise not)';
  CSTI_ShiftLeft: Result := '<< (left shift)';
  CSTI_ShiftRight: Result := '>> (right shift)';

  CSTI_Assignment_Plus: Result := '+=';
  CSTI_Assignment_Minus: Result := '-=';
  CSTI_Assignment_Divide: Result := '/=';
  CSTI_Assignment_Multiply: Result := '*=';

  CSTI_Assignment_LogicalOr: Result := '||= (bitwise or assignment)';
  CSTI_Assignment_LogicalXor: Result := '^^= (bitwise xor assignment)';
  CSTI_Assignment_LogicalAnd: Result := '&&= (bitwise and assignment)';

  CSTI_Assignment_BitwiseOr: Result := '|= (bitwise or assignment)';
  CSTI_Assignment_BitwiseXor: Result := '^= (bitwise xor assignment)';
  CSTI_Assignment_BitwiseAnd: Result := '&= (bitwise and assignment)';

  CSTI_Assignment_ShiftLeft: Result := '<<= (left shift assignment)';
  CSTI_Assignment_ShiftRight: Result := '>>= (right shift assignment)';

  CSTI_Integer: Result := 'integer data type';
  CSTI_HexInt : Result := 'hexadecimal integer data type';
  CSTI_Real: Result := 'floatpoint data type';
  CSTI_String: Result := 'string data type';

{Identifiers}
  CSTII_Begin: Result := '{';
  CSTII_End: Result := '}';

  CSTII_For: Result := 'for';
  CSTII_If: Result := 'if';
  CSTII_Else: Result := 'else';
  CSTII_While: Result := 'while';
  CSTII_do: Result := 'do';
  CSTII_Break: Result := 'break';
  CSTII_Continue: Result := 'continue';
  CSTII_Return : Result := 'return';

  CSTII_piece : result := 'piece';
  CSTII_GlobalVar: Result := 'static-var';
  CSTII_LocalVar: Result := 'var';
  CSTII_GetUnitProperty: Result := 'get';
  CSTII_SetUnitProperty: Result := 'set';
  CSTII_const: Result := 'const';

  CSTII_Sleep: Result := 'sleep';
  CSTII_WaitForTurn: Result := 'wait-for-turn';
  CSTII_WaitForMove: Result := 'wait-for-move';
  CSTII_Turn: Result := 'turn';
  CSTII_Move: Result := 'move';
  CSTII_Spin: Result := 'spin';
  CSTII_StopSpin: Result := 'stop-spin';
  CSTII_To: Result := 'to';
  CSTII_Around: Result := 'around';
  CSTII_along: Result := 'along';
  CSTII_from: Result := 'from';
  CSTII_Now: Result := 'now';
  CSTII_Speed: Result := 'speed';
  CSTII_Accelerate: Result := 'accelerate';
  CSTII_Decelerate: Result := 'decelerate';
  CSTII_Hide: Result := 'hide';
  CSTII_Show: Result := 'show';
  CSTII_Cache: Result := 'cache';
  CSTII_DontCache: Result := 'dont-cache';
  CSTII_CallScript: Result := 'call-script';
  CSTII_StartScript: Result := 'start-script';
  CSTII_Signal: Result := 'signal';
  CSTII_SetSignalMask: Result := 'set-signal-mask';
  CSTII_Explode: Result := 'explode';
  CSTII_ExplodeType: Result := 'type';
  CSTII_EmitSfx : result := 'emit-sfx';
  CSTII_Shade : result := 'shade';
  CSTII_DontShade : result := 'dont-shade';
  CSTII_rand : result := 'rand';
  CSTII_AttachUnit: result := 'attach-unit'; 
  CSTII_DropUnit: result := 'drop-unit';
  CSTII_BASMInline : result := 'BASM'
else Result := 'Unknown token'+IntToStr( ord( aToken ) );
end;
end; {TokenName}

function MakeKeywordTable : TStringlist;
var
  index : integer;
begin
result := TStringList.create;
result.sorted := true;
result.Duplicates := dupError;
result.Capacity := High(KeyWordlist) - low(KeyWordlist);
// biuld the keyword array
For index := low(KeyWordlist) to High(KeyWordlist) do
  result.AddObject(KeyWordlist[index].Symbol,TObject(KeyWordlist[index].Token));
end; {MakeKeywordTable}

// Recognize a Decimal Digit
function IsDigit(c: char): boolean;
begin
Result := c in ['0'..'9',HexadecimalChar];
end; {IsDigit}

// Recognize an Alpha Character
function IsAlpha(c: char): boolean;
begin
Result := UpCase(c) in ['A'..'Z','_'];
end; {IsAlpha}

// Recognize an AlphaNumeric Character
function IsAlNum(c: char): boolean;
begin
result := IsAlpha(c) or IsDigit(c);
end; {IsAlNum}

// Recognize newline Characters
function IsNewLine(c: char) : boolean;
begin
Result := c in [LF,CR];
end; {IsNewLine}

// Recognize Any Operator
function IsOp(c: char): boolean;
begin
Result := c in [ '+', '-', '*', '/', '<', '>', ':',
                 '=','.',';','(',')',',',';','[',']',
                 '{','}','!','|','&'];
end; {IsOp}

// -----------------------------------------------------------------------------

Constructor TScanner.Create( Const atext : string);
begin
inherited create;
fText := atext;
// get the keyword table
KeyWords := MakeKeywordTable;

// set up the input stream
fNextCharIndex := 0;
GetChar;
GetChar;  // pre-load the char to process
GetToken; // pre-load token
end; {Create}

Destructor TScanner.Destroy;
begin
FreeAndNil(KeyWords);
Inherited destroy;
end; {destroy}

Procedure TScanner.Error(Const s: string; ErrorAddress: Pointer);
begin
//WriteMsg('');
//WriteMsg('Error: '+ s+ '.');
{$IFDEF BreakOnError}
asm int 3 end;
{$ENDIF}
raise EScannerError.create('Error: '+ s+ '.') at ErrorAddress;
end; {Error}

procedure TScanner.Expected(Const s,a: string);
var ErrorAddress : Pointer;
begin
asm
  MOV eax,[EBP+4];
  MOV ErrorAddress,eax;
end;
Error('"'+s + '" expected but "' +a+ '" found',ErrorAddress);
end; {Expected}

procedure TScanner.Expected(Const s,a: string; ErrorAddress: Pointer);
begin
Error('"'+s + '" expected but "' +a+ '" found',ErrorAddress);
end; {Expected}

Procedure TScanner.Scanner_Abort(Const s : string);
var ErrorAddress : Pointer;
begin
// load the error data
fToken.Value := look;
fToken.Token := CSTINT_WhiteSpace;
fToken.Position := NextCharIndex-1;
// Report the error
asm
  MOV eax,[EBP+4];
  MOV ErrorAddress,eax;
end;
Error(s, ErrorAddress);
end; {Scanner_Abort}

// -----------------------------------------------------------------------------

// Read New Character From Input Stream
function TScanner.GetChar : char;
begin
look := NextLook;
if NextCharIndex < length(ftext) then
  begin
  inc(fNextCharIndex);
  NextLook := ftext[NextCharIndex];
  end
else
  NextLook := #0;
result := look;
end; {GetChar}

// Inserts a character into the text stream
Procedure TScanner.InsertString( index : integer;
                                 const substr : string;
                                 AutoSkipWhiteSpace: boolean = true );
begin
insert(substr, ftext, index);
if ( NextCharIndex = 0) or
   ( NextCharIndex = 1) and
   ( length(ftext) >= 1) then
  begin
  Look := ftext[1];
  if length(ftext) >= 2 then
    NextLook := ftext[2];
  end
else
  begin
  if (NextCharIndex > 0) and
     (NextCharIndex-1 < length(ftext)) then
    Look := ftext[NextCharIndex-1]
  else
    Look := #0;
  if (NextCharIndex > 0) and
     (NextCharIndex < length(ftext)) then
    NextLook := ftext[NextCharIndex]
  else
    NextLook := #0;
  end;
if AutoSkipWhiteSpace then
  SkipWhite;
end; {InsertString}

Procedure TScanner.RemoveString(index,count : integer);
begin
Delete(fText,index,count);
if NextCharIndex > index then
  fNextCharIndex := NextCharIndex - count;
if NextCharIndex > length(fText) then
  fNextCharIndex := length(fText);
if ( NextCharIndex = 0) or
   ( NextCharIndex = 1) and
   ( length(ftext) >= 1) then
  begin
  Look := ftext[1];
  if length(ftext) >= 2 then
    NextLook := ftext[2];
  fNextCharIndex := 1;  
  end
else
  begin
  if NextCharIndex-1 > 0 then
    Look := ftext[NextCharIndex-1]
  else
    Look := #0;
  if NextCharIndex > 0 then
    NextLook := ftext[NextCharIndex]
  else
    NextLook := #0;
  end;
SkipWhite;
end; {RemoveString}

// -----------------------------------------------------------------------------

// Skip Over a commented section, if 'ConvertCommentsToWhiteSpace' is defined,
// the comments are converted into a white space char
Procedure TScanner.SkipComment;
var
  count : Longword;
{$IFDEF ConvertCommentsToWhiteSpace}
  SkippedComment : boolean;
{$ENDIF}
begin
{$IFDEF ConvertCommentsToWhiteSpace}
SkippedComment := false;
{$ENDIF}
// comment type one
if (Look = CommentType1[1]) and (NextLook = CommentType1[2]) then
  begin
  while not IsNewLine(Look) and (look <> #0) do
    GetChar;
  if look = LF then
    begin
    GetChar;
    if look = CR then
      GetChar;
    end;
{$IFDEF ConvertCommentsToWhiteSpace}
  SkippedComment := true;
{$ENDIF}
  end;
// comment type 2
if (Look = CommentType2Start[1]) and (NextLook =CommentType2Start[2]) then
  begin
  count := 0;
  repeat
    if (Look = CommentType2Start[1]) and (NextLook =CommentType2Start[2]) then
      begin
      inc(count);
      GetChar;
      end
    else if (Look = CommentType2End[1]) and (NextLook =CommentType2End[2]) then
      begin
      dec(count);
      GetChar;
      end;
    GetChar;
    if look = #0 then
      Scanner_Abort('Unexpected end of file in comment');
  Until count = 0;
{$IFDEF ConvertCommentsToWhiteSpace}
  SkippedComment := true;
{$ENDIF}
  end;
{$IFDEF ConvertCommentsToWhiteSpace}
if SkippedComment then
  begin
  Look := ' ';
  dec( fNextCharIndex );
  if fNextCharIndex < 2 then
    fNextCharIndex := 2;
  if fNextCharIndex < Length( ftext ) then
    NextLook := ftext[ fNextCharIndex ]
  else
    NextLook := #0;
  end;
{$ENDIF}
end; {SkipComment}

// Skip Over Leading White Space
procedure TScanner.SkipWhite;
begin
SkipComment;
while Look in [' ', TAB, LF, CR] do
  begin
  GetChar;
  SkipComment;
  end;
end; {SkipWhite}

// Skip Over a Comma if it exists.
function TScanner.SkipComma : boolean;
begin
result := Token = CSTI_Comma;
if result then
  GetToken;
end; {SkipComma}

procedure TScanner.SkipToNewLine;
begin
while not IsNewLine(Look) and (Look <> #0) do
  GetChar;
end; {SkipToNewLine}

{
 Table Lookup
 If the input string matches a table entry, return the entry
 index.  If not, return a zero. (case-sensitive)
}
function TScanner.KeyWordLookup(Const Symbol : string): TToken;
var
  Index : integer;
begin
if (Symbol = '') or (length(symbol) = 1) then
  begin
  Result := CSTI_Identifier;
  exit;
  end;
if KeyWords.Find(Symbol,Index) then
  Result := TToken(KeyWords.Objects[index])
else  // unknown keyword, return its token type as an identifier
  Result := CSTI_Identifier;
end; {Lookup}

// Get an Identifier
Procedure TScanner.GetIdentifier;
var
  index : integer;
  StopLooping : boolean;
begin
fToken.Token := CSTI_Identifier;  
if not IsAlpha(Look) then
  Expected('Identifier',Look);
setLength(fToken.Value,8);
index := 0;
StopLooping := false;
repeat
  while IsAlNum(Look) do
    begin
    if look = HexadecimalChar then
      Scanner_Abort('Invalid char '''+HexadecimalChar+''' in identifier name');
    inc(index);
    if index >= length(fToken.Value) then
      setlength(fToken.Value,index+8);
    fToken.Value[index] := Look;
    GetChar;
    SkipComment;
    end;
  if (look = '-') and IsAlNum(NextLook) then
    begin
    GetChar;
    inc(index);
    if index >= length(fToken.Value) then
      setlength(fToken.Value,index+8);
    fToken.Value[index] := '-';  
    end
  else
    begin
    setLength(fToken.Value,index);
    fToken.Token := KeyWordLookup(fToken.Value);
    StopLooping := true;
    end
until StopLooping;
{$IFNDEF CaseSensitive}
fToken.Value := AnsiUpperCase( fToken.Value );
{$ENDIF}
end; {GetIdentifier}

// Get a Number
Procedure TScanner.GetNumber;
var
  index : integer;
  IsFloat : boolean;
begin
fToken.Token := CSTI_Integer;
if not IsDigit(Look) then
  Expected('Integer or floatpoint',Look);
setLength(ftoken.Value,16);
index := 0;
if look = HexadecimalChar then
  ftoken.Token := CSTI_HexInt;
IsFloat := false;
while IsDigit(Look) or (look = '.') do
  begin
  inc(index);
  if index >= length(ftoken.Value) then
    setlength(ftoken.Value,index+16);
  if look = '.' then
    begin
    if not IsFloat then
      begin
      ftoken.Token := CSTI_Real;
      IsFloat := true;
      end
    else if ftoken.Token = CSTI_HexInt then
      Scanner_Abort('Invalid number format, mixing Hexadecimal with floatpoint')
    else
      Scanner_Abort('Invalid floatpoint number, duplicate ''.''')
    end
  else if look = HexadecimalChar then
    Scanner_Abort('Invalid Hexadecimal number, duplicate '''+HexadecimalChar+'''');
  ftoken.Value[index] := Look;
  GetChar;
  SkipComment;
  end;
if index <> 0 then
  begin
  if IsFloat and (ftoken.value[index] = '.') then
    begin // demote the variable type
    ftoken.Token := CSTI_Integer;
    dec(index);
    // emit a warning
//    Hint(ftoken.position,'Floatpoint demoted to integer : '''+ftoken.value+'.''');
    end;
  Setlength(ftoken.Value,index);
  end
else
  ftoken.Value := ''; 
end; {GetNumber}

// Get a string
procedure TScanner.GetString;
var
  index : integer;
  tmp : string;
begin
fToken.Token := CSTI_String;
if (Look <> StringChar) then
  Expected('String character : '''+StringChar+'''',Look);
GetChar; // eat the start of string char;
setLength(tmp,8);
index := 0;
fToken.value := '';
Repeat
  // check to see if this is the end of the string or just imbedding a ' into the string
  // or it is the end of the string
  if (look = StringChar) then
    begin
    GetChar;
    if (look <> StringChar) then
      begin
      if index <> 0 then
        begin
        setLength(tmp,index);
        fToken.value := tmp;
        end;
      exit;
      end;
    GetChar;
    InsertString( NextCharIndex-1, StringChar );
    end;
  inc(index);
  if index > length(tmp) then
    setlength(tmp,index+8);
  tmp[index] := Look;  
  GetChar;
until IsNewLine(look) or (look = #0);
if IsNewLine(look) then
  Scanner_Abort('Unterminated string')
else if Look = #0 then // end of file
  Scanner_Abort('Unexpected end of file in string')
else
  Scanner_Abort('Error tokenizing string in lexical scanner');
end; {GetString}

Procedure TScanner.GetOperator;
var
  OldPos : integer;
  OldLook : char;
  OldNextLook : char;
  NotValid : boolean;
begin
ftoken.Token := CSTINT_WhiteSpace;
if not IsOp(Look) then
  Expected('Operator',Look);
// biuld the operator type
ftoken.value := look;
case Look of
  // single char operators
  ';' : ftoken.Token := CSTI_SemiColon;
  '(' : ftoken.Token := CSTI_OpenRound;
  ')' : ftoken.Token := CSTI_CloseRound;
  '.' : ftoken.Token := CSTI_Period;
  ',' : ftoken.Token := CSTI_Comma;
  ':' : ftoken.Token := CSTI_Colon;
  ']' : ftoken.Token := CSTI_CloseBlock;
  '{' : ftoken.Token := CSTII_Begin;
  '}' : ftoken.Token := CSTII_End;
  '~' : ftoken.Token := CSTI_LogicalNot;
end; {case}
if ftoken.Token <> CSTINT_WhiteSpace then
  begin // a token has been parsed
  GetChar;
  exit;
  end;
// go to the next posible token size
case Look of
  // multi sized operators
  '+' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_Plus;
      end
    else if NextLook = '+' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Increment;
      end
    else
      ftoken.Token := CSTI_Plus;
    end;
  '-' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_Minus;
      end
    else if NextLook = '-' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Decrement;
      end
    else
      ftoken.Token := CSTI_Minus;
    end;
  '*' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_Multiply;
      end
    else
      ftoken.Token := CSTI_Multiply;
    end;
  '/' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_Divide;
      end
    else
      ftoken.Token := CSTI_Divide;
    end;
  '|' :
    begin
    if NextLook = '|' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LogicalOr;
      if NextLook = '=' then
        begin
        ftoken.value := ftoken.value + GetChar;
        ftoken.Token := CSTI_Assignment_LogicalOr;
        end;
      end
    else if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_BitwiseOr;
      end
    else
      ftoken.Token := CSTI_BitwiseOr
    end;
  '&' :
    begin
    if NextLook = '&' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LogicalAnd;
      if NextLook = '=' then
        begin
        ftoken.value := ftoken.value + GetChar;
        GetChar;
        ftoken.Token := CSTI_Assignment_LogicalAnd;
        end;
      end
    else if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_BitwiseAnd;
      end
    else
      ftoken.Token := CSTI_BitwiseAnd;
    end;
  '^' :
    begin
    if NextLook = '^' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LogicalXor;
      if NextLook = '=' then
        begin
        ftoken.value := ftoken.value + GetChar;
        ftoken.Token := CSTI_Assignment_LogicalXor;
        end;
      end
    else if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Assignment_BitwiseXor;
      end
    else
      ftoken.Token := CSTI_BitwiseXor;
    end;
  '!' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_NotEqual;
      end
    else
      ftoken.Token := CSTI_BitwiseNot;
    end;
  '=' :
    begin
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_Equal;
      end
    else
      ftoken.Token := CSTI_Assignment;
    end;
  '[' :
    begin
    OldPos := NextCharIndex;
    OldLook := Look;
    OldNextLook := NextLook;
    NotValid := true;
    GetChar;
    SkipWhite;
    if IsDigit(Look) or (look = '-') then
      begin
      if (look = '-') then
        begin
        GetChar;
        SkipWhite;
        GetNumber;
        ftoken.Value := '-'+ftoken.Value;
        end
      else GetNumber;
      case ftoken.Token of
        CSTI_Integer : ftoken.Value := IntToStr( Round( StrToInt( ftoken.Value ) * LinearConst ));
        CSTI_Real :
          begin
          ftoken.Value := IntToStr( Round( StrToFloat( ftoken.Value ) * LinearConst ));
          ftoken.Token := CSTI_Integer;
          end;
        else Scanner_Abort('Internal error in TScanner.GetOperator(''['')');
      end;
      SkipWhite;
      NotValid := Look <> ']';
      end;
    if NotValid then
      begin
      fNextCharIndex := OldPos;
      Look := OldLook;
      NextLook := OldNextLook;
      ftoken.Token := CSTI_OpenBlock;
      end;
    end;
  '<' :
    if NextLook = '=' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_LessEqual;
      end
    else if NextLook = '<' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_ShiftLeft;
      if NextLook = '=' then
        begin
        ftoken.value := look + GetChar;
        ftoken.Token := CSTI_Assignment_ShiftLeft;
        end
      end
    else
      begin
      OldPos := NextCharIndex;
      OldLook := Look;
      OldNextLook := NextLook;
      NotValid := true;
      GetChar;
      SkipWhite;
      if IsDigit(Look) or (look = '-') then
        begin
        if (look = '-') then
          begin
          GetChar;
          SkipWhite;
          GetNumber;
          ftoken.Value := '-'+ftoken.Value;
          end
        else GetNumber;
        case ftoken.Token of
          CSTI_Integer : ftoken.Value := IntToStr( Round( StrToInt( ftoken.Value ) * AngularConst ));
          CSTI_Real :
            begin
            ftoken.Value := IntToStr( Round( StrToFloat( ftoken.Value ) * AngularConst ));
            ftoken.Token := CSTI_Integer;
            end;
        else Scanner_Abort('Internal error in TScanner.GetOperator(''<'')');
        end;
        SkipWhite;
        NotValid := Look <> '>';
        end;
      if NotValid then
        begin
        fNextCharIndex := OldPos;
        Look := OldLook;
        NextLook := OldNextLook;
        ftoken.Token := CSTI_Less;
        end;
      end;
  '>' :
    begin
    if (NextLook = '=') then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_GreaterEqual
      end
    else if NextLook = '>' then
      begin
      ftoken.value := look + GetChar;
      ftoken.Token := CSTI_ShiftRight;
      if NextLook = '=' then
        begin
        ftoken.value := look + GetChar;
        ftoken.Token := CSTI_Assignment_ShiftRight;
        end
      end
    else ftoken.Token := CSTI_Greater;
    end;
end; {case}
if ftoken.Token = CSTINT_WhiteSpace then
  Scanner_Abort('Unknown operator type : '''+ ftoken.value +'''');
GetChar;
end; {GetOperator}

Function TScanner.TillNewLine : string;
var
  len, count : integer;
begin
len := 16;
Setlength( Result, len );
count := 0;
while not IsNewLine(Look) and (look <> #0) do
  begin
  GetChar;
  inc( count );
  if count > len then
    begin
    len := count * 2;
    Setlength(Result, len);
    end;
  Result[Count] := Look;
  end;
while (count > 0) and IsNewLine(Result[count]) do
  dec(count);
setlength(Result,count);
end; {TillNewLine}

// Lexical Scanner
Procedure TScanner.GetToken;
var
  PrevToken : TTokenData;
  Done : boolean;
  NextTokenWasBlank : boolean;
begin
// change the position of the current token
NextTokenWasBlank := (NextToken.Token <> CSTINT_WhiteSpace) and
                     (Token <> CSTINT_WhiteSpace);
repeat
  PrevToken := NextToken;
  fToken := PrevToken;
  SkipWhite;
  SkipComment;
  fToken.Position := NextCharIndex-1;
  repeat
    Done := true;
    SkipWhite;
    SkipComment;
    if IsAlpha(Look) then
      GetIdentifier
    else if IsDigit(Look) then
      GetNumber
    else if look = StringChar then
      GetString
    else if Look = PreProcessorCommand then
      begin
      asm int 3 end; // should not get preprocessor commands!!
      end
    else if IsOp(Look) then
      GetOperator
    else if Look = #0 then
      begin
      fToken.Token := CSTI_EOF;
      fToken.Value := '';
      end
    else
      begin
      fToken.Value := Look;
      fToken.Token := CSTI_Identifier;
      //  asm int 3 end; // Unknown token!!!
      GetChar;
      end;
  until Done;
  // store the token it its correct spot
  fNextToken := fToken;
  fToken := PrevToken;
  if NextTokenWasBlank and (fToken.Token <> CSTINT_WhiteSpace) then
    NextTokenWasBlank := false;
until fToken.Token <> CSTINT_WhiteSpace;
end; {GetToken}


Procedure TScanner.Match( aToken : TToken );
var ErrorAddress : Pointer;
begin
if Token = aToken then
  GetToken
else
  begin
  asm
    MOV eax,[EBP+4];
    MOV ErrorAddress,eax;
  end;
  if Token = CSTI_Identifier then
    Expected(TokenName(aToken),Value,ErrorAddress)
  else
    Expected(TokenName(aToken),TokenName(Token),ErrorAddress);
  end;
end; {Match}

function TScanner.IfMatch( aToken : TToken ): boolean;
begin
if Token = CSTI_EOF then
  Match( aToken );
result := Token = aToken;
if result then
  GetToken;
end; {IfMatch}

end.
